from dataclasses import dataclass
from typing import Dict, Tuple

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem


@dataclass(frozen=True)
class ItemDatabase:
    major_items: Dict[str, MajorItem]
    ammo: Dict[str, Ammo]
    default_items: Dict[ItemCategory, Tuple[MajorItem, ...]]


def read_database(major_items_data: Dict,
                  ammo_data: Dict,
                  ) -> ItemDatabase:
    """

    :param major_items_data:
    :param ammo_data:
    :return:
    """
    major_items = {
        name: MajorItem.from_json(name, value)
        for name, value in major_items_data["items"].items()
    }

    ammo = {
        name: Ammo.from_json(name, value)
        for name, value in ammo_data.items()
    }

    default_items = {
        ItemCategory(category_name): tuple(major_items[item_name] for item_name in value)
        for category_name, value in major_items_data["default_items"].items()
    }

    return ItemDatabase(major_items, ammo, default_items)


def write_database(database: ItemDatabase,
                   ) -> Tuple[Dict, Dict]:
    """

    :param database:
    :return:
    """
    major_items_data = {
        name: item.as_json
        for name, item in database.major_items.items()
    }

    ammo_data = {
        name: ammo.as_json
        for name, ammo in database.ammo.items()
    }

    default_data = {
        category.value: [item.name for item in items]
        for category, items in database.default_items.items()
    }

    return {"items": major_items_data, "default_items": default_data}, ammo_data


_TEMPLE_KEYS = ["Dark Agon Key", "Dark Torvus Key", "Ing Hive Key"]


def add_memo_data_keys(data: dict):
    for i in range(1, 4):
        for temple_key in _TEMPLE_KEYS:
            data["{} {}".format(temple_key, i)] = data[temple_key]

    for temple_key in _TEMPLE_KEYS:
        data.pop(temple_key)

    for i in range(1, 10):
        data["Sky Temple Key {}".format(i)] = data["Sky Temple Key"]
    data.pop("Sky Temple Key")



from typing import Optional

import wiiload
from PySide2 import QtWidgets
from asyncqt import asyncSlot

from randovania import get_data_path
from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.game_connection.dolphin_backend import DolphinBackend
from randovania.game_connection.game_connection import GameConnection
from randovania.game_connection.nintendont_backend import NintendontBackend
from randovania.gui.lib import async_dialog
from randovania.interface_common.options import Options, InfoAlert


class GameConnectionSetup:
    def __init__(self, parent: QtWidgets.QWidget, tool_button: QtWidgets.QToolButton,
                 label: QtWidgets.QLabel, connection: GameConnection, options: Options):
        self.parent = parent
        self.tool = tool_button
        self.label = label
        self.game_connection = connection
        self.options = options

        self.game_connection.Updated.connect(self.on_game_connection_updated)
        self.tool.setText("Configure backend")

        def _create_check(text: str, on_triggered, default: Optional[bool] = None):
            action = QtWidgets.QAction(self.tool)
            action.setText(text)
            action.setCheckable(True)
            if default is not None:
                action.setChecked(default)
            action.triggered.connect(on_triggered)
            return action

        self.game_connection_menu = QtWidgets.QMenu(self.tool)

        self.use_dolphin_backend = _create_check("Dolphin", self.on_use_dolphin_backend)
        self.use_nintendont_backend = _create_check("", self.on_use_nintendont_backend)
        self.upload_nintendont_action = QtWidgets.QAction(self.tool)
        self.upload_nintendont_action.setText("Upload Nintendont to Homebrew Channel")
        self.upload_nintendont_action.triggered.connect(self.on_upload_nintendont_action)
        self.connect_to_game = _create_check("Connect to the game", self.on_connect_to_game, True)

        self.game_connection_menu.addAction(self.use_dolphin_backend)
        self.game_connection_menu.addAction(self.use_nintendont_backend)
        self.game_connection_menu.addAction(self.upload_nintendont_action)
        self.game_connection_menu.addSeparator()
        self.game_connection_menu.addAction(self.connect_to_game)

        self.tool.setMenu(self.game_connection_menu)

        self.refresh_backend()
        self.on_game_connection_updated()

    def on_game_connection_updated(self):
        self.label.setText(self.game_connection.pretty_current_status)

    def refresh_backend(self):
        backend = self.game_connection.backend

        self.use_dolphin_backend.setChecked(isinstance(backend, DolphinBackend))
        if isinstance(backend, NintendontBackend):
            self.use_nintendont_backend.setChecked(True)
            self.use_nintendont_backend.setText(f"Nintendont: {backend.ip}")
            self.upload_nintendont_action.setEnabled(True)
        else:
            self.use_nintendont_backend.setChecked(False)
            self.use_nintendont_backend.setText(f"Nintendont")
            self.upload_nintendont_action.setEnabled(False)

    def on_use_dolphin_backend(self):
        self.game_connection.set_backend(DolphinBackend())
        with self.options as options:
            options.game_backend = GameBackendChoice.DOLPHIN
        self.refresh_backend()

    @asyncSlot()
    async def on_use_nintendont_backend(self):
        dialog = QtWidgets.QInputDialog(self.parent)
        dialog.setModal(True)
        dialog.setWindowTitle("Enter Wii's IP")
        dialog.setLabelText("Enter the IP address of your Wii. "
                            "You can check the IP address on the pause screen of Homebrew Channel.")
        if self.options.nintendont_ip is not None:
            dialog.setTextValue(self.options.nintendont_ip)

        if await async_dialog.execute_dialog(dialog) == dialog.Accepted:
            new_ip = dialog.textValue()
            if new_ip != "":
                if not self.options.is_alert_displayed(InfoAlert.NINTENDONT_UNSTABLE):
                    await async_dialog.warning(self.parent, "Nintendont Limitation",
                                               "Warning: The Nintendont integration isn't perfect and is known to "
                                               "crash.")
                    self.options.mark_alert_as_displayed(InfoAlert.NINTENDONT_UNSTABLE)

                with self.options as options:
                    options.nintendont_ip = new_ip
                    options.game_backend = GameBackendChoice.NINTENDONT
                self.game_connection.set_backend(NintendontBackend(new_ip))
        self.refresh_backend()

    @asyncSlot()
    async def on_upload_nintendont_action(self):
        nintendont_file = get_data_path().joinpath("nintendont", "boot.dol")
        if not nintendont_file.is_file():
            return await async_dialog.warning(self.parent, "Missing Nintendont",
                                              "Unable to find a Nintendont executable.")

        text = f"Uploading Nintendont to the Wii at {self.options.nintendont_ip}..."
        box = QtWidgets.QMessageBox(QtWidgets.QMessageBox.NoIcon, "Uploading to Homebrew Channel",
                                    text, QtWidgets.QMessageBox.Ok, self.parent)
        box.button(QtWidgets.QMessageBox.Ok).setEnabled(False)
        box.show()

        try:
            await wiiload.upload_file(nintendont_file, [], self.options.nintendont_ip)
            box.setText(f"Upload finished successfully. Check your Wii for more.")
        except Exception as e:
            box.setText(f"Error uploading to Wii: {e}")
        finally:
            box.button(QtWidgets.QMessageBox.Ok).setEnabled(True)

    def on_connect_to_game(self):
        connect_to_game = self.connect_to_game.isChecked()
        self.game_connection.backend.set_connection_enabled(connect_to_game)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trick_details_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_TrickDetailsPopup(object):
    def setupUi(self, TrickDetailsPopup):
        if not TrickDetailsPopup.objectName():
            TrickDetailsPopup.setObjectName(u"TrickDetailsPopup")
        TrickDetailsPopup.resize(416, 337)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(TrickDetailsPopup.sizePolicy().hasHeightForWidth())
        TrickDetailsPopup.setSizePolicy(sizePolicy)
        TrickDetailsPopup.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(TrickDetailsPopup)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.title_label = QLabel(TrickDetailsPopup)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setWordWrap(True)
        self.title_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.title_label)

        self.scroll_area = QScrollArea(TrickDetailsPopup)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 396, 191))
        self.verticalLayout_2 = QVBoxLayout(self.scroll_area_contents)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.area_list_label = QLabel(self.scroll_area_contents)
        self.area_list_label.setObjectName(u"area_list_label")
        self.area_list_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.area_list_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.area_list_label)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        self.button_box = QDialogButtonBox(TrickDetailsPopup)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Close)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(TrickDetailsPopup)

        QMetaObject.connectSlotsByName(TrickDetailsPopup)
    # setupUi

    def retranslateUi(self, TrickDetailsPopup):
        TrickDetailsPopup.setWindowTitle(QCoreApplication.translate("TrickDetailsPopup", u"Item Configuration", None))
        self.title_label.setText(QCoreApplication.translate("TrickDetailsPopup", u"<html><head/><body><p><span style=\" font-weight:600;\">{trick.long_name}</span></p><p>{trick.description}</p><p>{trick.long_name} with level {level} can be used in the rooms listed below.</p><p>Click a room name to open it in the Data Visualizer for more details.</p></body></html>", None))
        self.area_list_label.setText(QCoreApplication.translate("TrickDetailsPopup", u"<html><head/><body><p><a href=\"data-editor://World/Area\">INSERT AREAS!</a></p></body></html>", None))
    # retranslateUi




import dataclasses
import struct
from enum import Enum

from randovania.dol_patching.assembler.ppc import *
from randovania.dol_patching.dol_file import DolFile
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.prime.all_prime_dol_patches import BasePrimeDolVersion
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class BeamCostAddresses:
    uncharged_cost: int
    charged_cost: int
    charge_combo_ammo_cost: int
    charge_combo_missile_cost: int
    get_beam_ammo_type_and_costs: int


@dataclasses.dataclass(frozen=True)
class SafeZoneAddresses:
    heal_per_frame_constant: int
    increment_health_fmr: int


@dataclasses.dataclass(frozen=True)
class StartingBeamVisorAddresses:
    player_state_constructor_clean: int
    player_state_constructor_decode: int
    health_info_constructor: int
    enter_morph_ball_state: int
    start_transition_to_visor: int


_PREFERENCES_ORDER = (
    "sound_mode",
    "screen_brightness",
    "screen_x_offset",
    "screen_y_offset",
    "screen_stretch",
    "sfx_volume",
    "music_volume",
    "hud_alpha",
    "helmet_alpha",
)
_FLAGS_ORDER = (
    "hud_lag",
    "invert_y_axis",
    "rumble",
    None,  # crashes, maybe swapBeamsControls
    "hint_system",
    None,  # 5: doesn't crash, unknown
    None,  # 6: doesn't crash, unknown
    None,  # 7: doesn't crash, unknown
)


def apply_game_options_patch(game_options_constructor_offset: int, user_preferences: EchoesUserPreferences,
                             dol_file: DolFile):
    patch = [
        # Unknown purpose, but keep for safety
        stw(r31, 0x1c, r1),
        or_(r31, r3, r3),

        # For a later function call we don't touch
        addi(r3, r1, 0x8),
    ]

    for i, preference_name in enumerate(_PREFERENCES_ORDER):
        value = getattr(user_preferences, preference_name)
        if isinstance(value, Enum):
            value = value.value
        patch.extend([
            li(r0, value),
            stw(r0, (0x04 * i), r31),
        ])

    flag_values = [
        getattr(user_preferences, flag_name)
        if flag_name is not None else False
        for flag_name in _FLAGS_ORDER
    ]
    bit_mask = int("".join(str(int(flag)) for flag in flag_values), 2)
    patch.extend([
        li(r0, bit_mask),
        stb(r0, 0x04 * len(_PREFERENCES_ORDER), r31),
        li(r0, 0),
        stw(r0, 0x2c, r31),
        stw(r0, 0x30, r31),
        stw(r0, 0x34, r31),
    ])

    instructions_space = 34
    instructions_to_fill = instructions_space - len(patch)

    if instructions_to_fill < 0:
        raise RuntimeError(f"Our patch ({len(patch)}) is bigger than the space we have ({instructions_space}).")

    for i in range(instructions_to_fill):
        patch.append(nop())
    dol_file.write_instructions(game_options_constructor_offset + 8 * 4, patch)


def apply_beam_cost_patch(patch_addresses: BeamCostAddresses, game_specific: EchoesGameSpecific,
                          dol_file: DolFile):
    uncharged_costs = []
    charged_costs = []
    combo_costs = []
    missile_costs = []
    ammo_types = []

    for beam_config in game_specific.beam_configurations:
        uncharged_costs.append(beam_config.uncharged_cost)
        charged_costs.append(beam_config.charged_cost)
        combo_costs.append(beam_config.combo_ammo_cost)
        missile_costs.append(beam_config.combo_missile_cost)
        ammo_types.append((
            beam_config.ammo_a.index if beam_config.ammo_a is not None else -1,
            beam_config.ammo_b.index if beam_config.ammo_b is not None else -1,
        ))

    uncharged_costs_patch = struct.pack(">llll", *uncharged_costs)
    charged_costs_patch = struct.pack(">llll", *charged_costs)
    combo_costs_patch = struct.pack(">llll", *combo_costs)
    missile_costs_patch = struct.pack(">llll", *missile_costs)

    # TODO: patch CPlayerGun::IsOutOfAmmoToShoot

    # The following patch also changes the fact that the game doesn't check if there's enough ammo for Power Beam
    # we start our patch right after the `addi r3,r31,0x0`
    ammo_type_patch_offset = 0x40
    ammo_type_patch = [
        0x81, 0x59, 0x07, 0x74,  # lwz r10,0x774(r25)           # r10 = get current beam
        0x55, 0x4a, 0x10, 0x3a,  # rlwinm r10,r10,0x2,0x0,0x1d  # r10 *= 4

        0x7c, 0x03, 0x50, 0x2e,  # lwzx r0,r3,r10               # r0 = BeamIdToUnchargedShotAmmoCost[currentBeam]
        *stw(r0, 0x0, r29).bytes_for(0),  # *outBeamAmmoCost = r0

        *lwz(r10, 0x774, r25).bytes_for(0),  # r10 = get current beam
        *addi(r10, r10, 0x1).bytes_for(0),  # r10 = r10 + 1
        0x7d, 0x49, 0x03, 0xa6,  # mtspr CTR,r10                # count_register = r10

        # Power Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz dark_beam               # if (--count_register > 0) goto
        *li(r3, ammo_types[0][0]).bytes_for(0),
        *li(r9, ammo_types[0][1]).bytes_for(0),
        0x42, 0x80, 0x00, 0x2c,  # b update_out_beam_type

        # Dark Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz dark_beam               # if (--count_register > 0) goto
        *li(r3, ammo_types[1][0]).bytes_for(0),
        *li(r9, ammo_types[1][1]).bytes_for(0),
        0x42, 0x80, 0x00, 0x1c,  # b update_out_beam_type

        # Light Beam
        0x42, 0x00, 0x00, 0x10,  # bdnz light_beam               # if (--count_register > 0) goto
        *li(r3, ammo_types[2][0]).bytes_for(0),
        *li(r9, ammo_types[2][1]).bytes_for(0),
        0x42, 0x80, 0x00, 0x0c,  # b update_out_beam_type

        # Annihilator Beam
        *li(r3, ammo_types[3][0]).bytes_for(0),
        *li(r9, ammo_types[3][1]).bytes_for(0),

        # update_out_beam_type
        *stw(r3, 0x0, r27).bytes_for(0),  # *outBeamAmmoTypeA = r3
        *stw(r9, 0x0, r28).bytes_for(0),  # *outBeamAmmoTypeB = r9

        0x42, 0x80, 0x00, 0x18,  # b body_end
        # jump to the code for getting the charged/combo costs and then check if has ammo
        # The address in question is at 0x801ccd64 for NTSC
    ]

    dol_file.write(patch_addresses.uncharged_cost, uncharged_costs_patch)
    dol_file.write(patch_addresses.charged_cost, charged_costs_patch)
    dol_file.write(patch_addresses.charge_combo_ammo_cost, combo_costs_patch)
    dol_file.write(patch_addresses.charge_combo_missile_cost, missile_costs_patch)
    dol_file.write(patch_addresses.get_beam_ammo_type_and_costs + ammo_type_patch_offset, ammo_type_patch)


def apply_safe_zone_heal_patch(patch_addresses: SafeZoneAddresses,
                               sda2_base: int,
                               game_specific: EchoesGameSpecific,
                               dol_file: DolFile):
    heal_per_second = game_specific.safe_zone_heal_per_second
    offset = patch_addresses.heal_per_frame_constant - sda2_base

    dol_file.write(patch_addresses.heal_per_frame_constant, struct.pack(">f", heal_per_second / 60))
    dol_file.write_instructions(patch_addresses.increment_health_fmr, [lfs(f1, offset, r2)])


def apply_starting_visor_patch(addresses: StartingBeamVisorAddresses, default_items: dict, dol_file: DolFile):
    visor_order = ["Combat Visor", "Echo Visor", "Scan Visor", "Dark Visor"]
    beam_order = ["Power Beam", "Dark Beam", "Light Beam", "Annihilator Beam"]

    default_visor = visor_order.index(default_items["visor"])
    default_beam = beam_order.index(default_items["beam"])

    # Patch CPlayerState constructor with default values
    dol_file.write_instructions(addresses.player_state_constructor_clean + 0x54, [
        bl(addresses.health_info_constructor),

        li(r0, default_beam),
        stw(r0, 0xc, r30),  # xc_currentBeam

        li(r0, default_visor),
        stw(r0, 0x30, r30),  # x30_currentVisor
        stw(r0, 0x34, r30),  # x34_transitioningVisor

        li(r3, 0),
    ])

    # Patch CPlayerState constructor for loading save files
    dol_file.write_instructions(addresses.player_state_constructor_decode + 0x5C, [
        li(r0, default_visor),
        stw(r0, 0x30, r30),
        stw(r0, 0x34, r30),
    ])

    # Patch EnterMorphBallState's unconditional call for StartTransitionToVisor, but only if default visor isn't combat
    if default_visor == 0:
        patch = [bl(addresses.start_transition_to_visor)]
    else:
        patch = [nop()]
    dol_file.write_instructions(addresses.enter_morph_ball_state + 0xEC, patch)


@dataclasses.dataclass(frozen=True)
class EchoesDolVersion(BasePrimeDolVersion):
    beam_cost_addresses: BeamCostAddresses
    safe_zone: SafeZoneAddresses
    starting_beam_visor: StartingBeamVisorAddresses



import copy
from typing import Iterator, Optional, Set, Dict, List, NamedTuple, Tuple

from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node, ResourceNode, PickupNode
from randovania.game_description.requirements import RequirementSet, Requirement, RequirementAnd, \
    ResourceRequirement
from randovania.generator import graph as graph_module
from randovania.resolver.state import State


class GraphPath(NamedTuple):
    previous_node: Optional[Node]
    node: Node
    requirement: Requirement

    def is_in_graph(self, digraph: graph_module.BaseGraph):
        if self.previous_node is None:
            return False
        else:
            return digraph.has_edge(self.previous_node.index, self.node.index)

    def add_to_graph(self, digraph: graph_module.BaseGraph):
        digraph.add_node(self.node.index)
        if self.previous_node is not None:
            digraph.add_edge(self.previous_node.index, self.node.index, requirement=self.requirement)


def filter_resource_nodes(nodes: Iterator[Node]) -> Iterator[ResourceNode]:
    for node in nodes:
        if node.is_resource_node:
            yield node


def filter_pickup_nodes(nodes: Iterator[Node]) -> Iterator[PickupNode]:
    for node in nodes:
        if isinstance(node, PickupNode):
            yield node


def filter_collectable(resource_nodes: Iterator[ResourceNode], reach: "GeneratorReach") -> Iterator[ResourceNode]:
    for resource_node in resource_nodes:
        if resource_node.can_collect(reach.state.patches, reach.state.resources, reach.game.world_list.all_nodes):
            yield resource_node


def filter_reachable(nodes: Iterator[Node], reach: "GeneratorReach") -> Iterator[Node]:
    for node in nodes:
        if reach.is_reachable_node(node):
            yield node


def filter_out_dangerous_actions(resource_nodes: Iterator[ResourceNode],
                                 game: GameDescription,
                                 ) -> Iterator[ResourceNode]:
    for resource_node in resource_nodes:
        if resource_node.resource() not in game.dangerous_resources:
            yield resource_node


class GeneratorReach:
    _digraph: graph_module.BaseGraph
    _state: State
    _game: GameDescription
    _reachable_paths: Optional[Dict[int, List[Node]]]
    _reachable_costs: Optional[Dict[int, int]]
    _node_reachable_cache: Dict[int, bool]
    _unreachable_paths: Dict[Tuple[Node, Node], Requirement]
    _safe_nodes: Optional[Set[int]]
    _is_node_safe_cache: Dict[Node, bool]

    def __deepcopy__(self, memodict):
        reach = GeneratorReach(
            self._game,
            self._state,
            self._digraph.copy()
        )
        reach._unreachable_paths = copy.copy(self._unreachable_paths)
        reach._reachable_paths = self._reachable_paths
        reach._reachable_costs = self._reachable_costs
        reach._safe_nodes = self._safe_nodes

        reach._node_reachable_cache = copy.copy(self._node_reachable_cache)
        reach._is_node_safe_cache = copy.copy(self._is_node_safe_cache)
        return reach

    def __init__(self,
                 game: GameDescription,
                 state: State,
                 graph: graph_module.BaseGraph
                 ):

        self._game = game
        self._state = state
        self._digraph = graph
        self._unreachable_paths = {}
        self._reachable_paths = None
        self._node_reachable_cache = {}
        self._is_node_safe_cache = {}

    @classmethod
    def reach_from_state(cls,
                         game: GameDescription,
                         initial_state: State,
                         ) -> "GeneratorReach":

        reach = cls(game, initial_state, graph_module.RandovaniaGraph.new())
        reach._expand_graph([GraphPath(None, initial_state.node, Requirement.trivial())])
        return reach

    def _potential_nodes_from(self, node: Node) -> Iterator[Tuple[Node, Requirement, bool]]:
        extra_requirement = _extra_requirement_for_node(self._game, node)
        requirement_to_leave = node.requirement_to_leave(self._state.patches, self._state.resources)

        for target_node, requirement in self._game.world_list.potential_nodes_from(node, self.state.patches):
            if target_node is None:
                continue

            if requirement_to_leave != Requirement.trivial():
                requirement = RequirementAnd([requirement, requirement_to_leave])

            if extra_requirement is not None:
                requirement = RequirementAnd([requirement, extra_requirement])

            satisfied = requirement.satisfied(self._state.resources, self._state.energy)
            yield target_node, requirement, satisfied

    def _expand_graph(self, paths_to_check: List[GraphPath]):
        # print("!! _expand_graph", len(paths_to_check))
        self._reachable_paths = None
        while paths_to_check:
            path = paths_to_check.pop(0)

            if path.is_in_graph(self._digraph):
                continue

            path.add_to_graph(self._digraph)

            for target_node, requirement, satisfied in self._potential_nodes_from(path.node):
                if satisfied:
                    paths_to_check.append(GraphPath(path.node, target_node, requirement))
                else:
                    self._unreachable_paths[path.node, target_node] = requirement

        self._safe_nodes = None

    def _can_advance(self,
                     node: Node,
                     ) -> bool:
        """
        Calculates if we can advance past a given node
        :param node:
        :return:
        """
        # We can't advance past a resource node if we haven't collected it
        if node.is_resource_node:
            return self._state.has_resource(node.resource())
        else:
            return True

    def _calculate_safe_nodes(self):
        if self._safe_nodes is not None:
            return

        for component in self._digraph.strongly_connected_components():
            if self._state.node.index in component:
                assert self._safe_nodes is None
                self._safe_nodes = component

        assert self._safe_nodes is not None

    def _calculate_reachable_paths(self):
        if self._reachable_paths is not None:
            return

        all_nodes = self.game.world_list.all_nodes

        def weight(source: int, target: int, attributes):
            if self._can_advance(all_nodes[target]):
                return 0
            else:
                return 1

        self._reachable_costs, self._reachable_paths = self._digraph.multi_source_dijkstra({self.state.node.index},
                                                                                           weight=weight)

    def is_reachable_node(self, node: Node) -> bool:
        index = node.index

        cached_value = self._node_reachable_cache.get(index)
        if cached_value is not None:
            return cached_value

        self._calculate_reachable_paths()

        cost = self._reachable_costs.get(index)
        if cost is not None:
            if cost == 0:
                self._node_reachable_cache[index] = True
            elif cost == 1:
                self._node_reachable_cache[index] = (not self._can_advance(node))
            else:
                self._node_reachable_cache[index] = False

            return self._node_reachable_cache[index]
        else:
            return False

    @property
    def connected_nodes(self) -> Iterator[Node]:
        """
        An iterator of all nodes there's an path from the reach's starting point. Similar to is_reachable_node
        :return:
        """
        self._calculate_reachable_paths()
        all_nodes = self.game.world_list.all_nodes
        for index in self._reachable_paths.keys():
            yield all_nodes[index]

    @property
    def state(self) -> State:
        return self._state

    @property
    def game(self) -> GameDescription:
        return self._game

    @property
    def nodes(self) -> Iterator[Node]:
        for node in self.game.world_list.all_nodes:
            if node.index in self._digraph:
                yield node

    @property
    def safe_nodes(self) -> Iterator[Node]:
        for node in self.nodes:
            if self.is_safe_node(node):
                yield node

    def is_safe_node(self, node: Node) -> bool:
        is_safe = self._is_node_safe_cache.get(node)
        if is_safe is not None:
            return is_safe

        self._calculate_safe_nodes()
        self._is_node_safe_cache[node] = node.index in self._safe_nodes
        return self._is_node_safe_cache[node]

    def advance_to(self, new_state: State,
                   is_safe: bool = False,
                   ) -> None:
        assert new_state.previous_state == self.state
        # assert self.is_reachable_node(new_state.node)

        if is_safe or self.is_safe_node(new_state.node):
            for index, _ in list(filter(lambda x: not x[1], self._node_reachable_cache.items())):
                del self._node_reachable_cache[index]

            for node, _ in list(filter(lambda x: not x[1], self._is_node_safe_cache.items())):
                del self._is_node_safe_cache[node]
        else:
            self._node_reachable_cache = {}
            self._is_node_safe_cache = {}

        self._state = new_state

        paths_to_check: List[GraphPath] = []

        edges_to_remove = []
        # Check if we can expand the corners of our graph
        # TODO: check if expensive. We filter by only nodes that depends on a new resource
        for edge, requirement in self._unreachable_paths.items():
            if requirement.satisfied(self._state.resources, self._state.energy):
                from_node, to_node = edge
                paths_to_check.append(GraphPath(from_node, to_node, requirement))
                edges_to_remove.append(edge)

        for edge in edges_to_remove:
            del self._unreachable_paths[edge]

        self._expand_graph(paths_to_check)

    def act_on(self, node: ResourceNode) -> None:
        all_nodes = self.game.world_list.all_nodes
        new_dangerous_resources = set(
            resource
            for resource, quantity in node.resource_gain_on_collect(self.state.patches, self.state.resources, all_nodes)
            if resource in self.game.dangerous_resources
        )
        new_state = self.state.act_on_node(node)

        if new_dangerous_resources:
            edges_to_remove = []
            for source, target, requirement in self._digraph.edges_data():
                # FIXME: don't convert to set!
                dangerous = requirement.as_set.dangerous_resources
                if dangerous and new_dangerous_resources.intersection(dangerous):
                    if not requirement.satisfied(new_state.resources, new_state.energy):
                        edges_to_remove.append((source, target))

            for edge in edges_to_remove:
                self._digraph.remove_edge(*edge)

        self.advance_to(new_state)

    def shortest_path_from(self, node: Node) -> Dict[Node, Tuple[Node, ...]]:
        if node.index in self._digraph:
            return self._digraph.single_source_dijkstra_path(node.index)
        else:
            return {}

    def unreachable_nodes_with_requirements(self) -> Dict[Node, RequirementSet]:
        results = {}
        for (_, node), requirement in self._unreachable_paths.items():
            if self.is_reachable_node(node):
                continue
            requirements = requirement.patch_requirements(self.state.resources, 1).simplify().as_set
            if node in results:
                results[node] = results[node].expand_alternatives(requirements)
            else:
                results[node] = requirement.as_set
        return results


def _extra_requirement_for_node(game: GameDescription, node: Node) -> Optional[Requirement]:
    extra_requirement = None

    if node.is_resource_node:
        resource_node: ResourceNode = node

        node_resource = resource_node.resource()
        if node_resource in game.dangerous_resources:
            extra_requirement = ResourceRequirement(node_resource, 1, False)

    return extra_requirement


def get_safe_resources(reach: GeneratorReach) -> Iterator[ResourceNode]:
    generator = filter_reachable(
        filter_out_dangerous_actions(
            collectable_resource_nodes(reach.nodes, reach),
            reach.game),
        reach
    )
    for node in generator:
        if reach.is_safe_node(node):
            yield node


def collectable_resource_nodes(nodes: Iterator[Node], reach: GeneratorReach) -> Iterator[ResourceNode]:
    return filter_collectable(filter_resource_nodes(nodes), reach)


def get_collectable_resource_nodes_of_reach(reach: GeneratorReach) -> List[ResourceNode]:
    return list(collectable_resource_nodes(filter_reachable(reach.nodes, reach), reach))


def collect_all_safe_resources_in_reach(reach: GeneratorReach) -> None:
    """

    :param reach:
    :return:
    """
    while True:
        actions = list(get_safe_resources(reach))
        if not actions:
            break

        for action in actions:
            if action.can_collect(reach.state.patches, reach.state.resources, reach.game.world_list.all_nodes):
                # assert reach.is_safe_node(action)
                reach.advance_to(reach.state.act_on_node(action), is_safe=True)


def reach_with_all_safe_resources(game: GameDescription, initial_state: State) -> GeneratorReach:
    """
    Creates a new GeneratorReach using the given state and then collect all safe resources
    :param game:
    :param initial_state:
    :return:
    """
    reach = GeneratorReach.reach_from_state(game, initial_state)
    collect_all_safe_resources_in_reach(reach)
    return reach


def advance_reach_with_possible_unsafe_resources(previous_reach: GeneratorReach) -> GeneratorReach:
    """
    Create a new GeneratorReach that collected actions not considered safe, but expanded the safe_nodes set
    :param previous_reach:
    :return:
    """

    game = previous_reach.game
    collect_all_safe_resources_in_reach(previous_reach)
    initial_state = previous_reach.state

    previous_safe_nodes = set(previous_reach.safe_nodes)

    for action in get_collectable_resource_nodes_of_reach(previous_reach):
        # print("Trying to collect {} and it's not dangerous. Copying...".format(action.name))
        next_reach = copy.deepcopy(previous_reach)
        next_reach.act_on(action)
        collect_all_safe_resources_in_reach(next_reach)

        if previous_safe_nodes <= set(next_reach.safe_nodes):
            # print("Non-safe {} was good".format(logic.game.node_name(action)))
            return advance_reach_with_possible_unsafe_resources(next_reach)

        if next_reach.is_reachable_node(initial_state.node):
            next_next_state = next_reach.state.copy()
            next_next_state.node = initial_state.node

            next_reach = reach_with_all_safe_resources(game, next_next_state)
            if previous_safe_nodes <= set(next_reach.safe_nodes):
                # print("Non-safe {} could reach back to where we were".format(logic.game.node_name(action)))
                return advance_reach_with_possible_unsafe_resources(next_reach)
        else:
            pass

    # We couldn't improve this reach, so just return it
    return previous_reach


def pickup_nodes_that_can_reach(pickup_nodes: Iterator[PickupNode],
                                reach: GeneratorReach,
                                safe_nodes: Set[Node]) -> Iterator[PickupNode]:
    for pickup_node in pickup_nodes:
        if pickup_node in safe_nodes or set(reach.shortest_path_from(pickup_node).keys()).intersection(safe_nodes):
            yield pickup_node


def advance_to_with_reach_copy(base_reach: GeneratorReach, state: State) -> GeneratorReach:
    """
    Copies the given Reach, advances to the given State and collect all possible resources.
    :param base_reach:
    :param state:
    :return:
    """
    potential_reach = copy.deepcopy(base_reach)
    potential_reach.advance_to(state)
    collect_all_safe_resources_in_reach(potential_reach)
    return potential_reach
    # return advance_reach_with_possible_unsafe_resources(potential_reach)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'tracker_window.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_TrackerWindow(object):
    def setupUi(self, TrackerWindow):
        if not TrackerWindow.objectName():
            TrackerWindow.setObjectName(u"TrackerWindow")
        TrackerWindow.resize(1040, 495)
        self.actionAction = QAction(TrackerWindow)
        self.actionAction.setObjectName(u"actionAction")
        self.menu_reset_action = QAction(TrackerWindow)
        self.menu_reset_action.setObjectName(u"menu_reset_action")
        self.centralWidget = QWidget(TrackerWindow)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout = QGridLayout(self.centralWidget)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.configuration_label = QLabel(self.centralWidget)
        self.configuration_label.setObjectName(u"configuration_label")
        self.configuration_label.setWordWrap(True)

        self.gridLayout.addWidget(self.configuration_label, 0, 0, 1, 3)

        self.tab_widget = QTabWidget(self.centralWidget)
        self.tab_widget.setObjectName(u"tab_widget")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tab_widget.sizePolicy().hasHeightForWidth())
        self.tab_widget.setSizePolicy(sizePolicy)
        self.tab_widget.setMinimumSize(QSize(290, 0))
        self.tab_widget.setMaximumSize(QSize(290, 16777215))
        self.tab_inventory = QWidget()
        self.tab_inventory.setObjectName(u"tab_inventory")
        self.tab_inventory_layout = QVBoxLayout(self.tab_inventory)
        self.tab_inventory_layout.setSpacing(0)
        self.tab_inventory_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_inventory_layout.setObjectName(u"tab_inventory_layout")
        self.tab_inventory_layout.setContentsMargins(0, 0, 0, 0)
        self.pickups_scroll_area = QScrollArea(self.tab_inventory)
        self.pickups_scroll_area.setObjectName(u"pickups_scroll_area")
        sizePolicy.setHeightForWidth(self.pickups_scroll_area.sizePolicy().hasHeightForWidth())
        self.pickups_scroll_area.setSizePolicy(sizePolicy)
        self.pickups_scroll_area.setMinimumSize(QSize(285, 0))
        self.pickups_scroll_area.setFrameShape(QFrame.NoFrame)
        self.pickups_scroll_area.setFrameShadow(QFrame.Plain)
        self.pickups_scroll_area.setLineWidth(0)
        self.pickups_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.pickups_scroll_area.setWidgetResizable(True)
        self.pickups_scroll_contents = QWidget()
        self.pickups_scroll_contents.setObjectName(u"pickups_scroll_contents")
        self.pickups_scroll_contents.setGeometry(QRect(0, 0, 285, 409))
        self.pickup_box_layout = QVBoxLayout(self.pickups_scroll_contents)
        self.pickup_box_layout.setSpacing(0)
        self.pickup_box_layout.setContentsMargins(11, 11, 11, 11)
        self.pickup_box_layout.setObjectName(u"pickup_box_layout")
        self.pickup_box_layout.setContentsMargins(0, 0, 0, 0)
        self.upgrades_box = QGroupBox(self.pickups_scroll_contents)
        self.upgrades_box.setObjectName(u"upgrades_box")
        self.upgrades_layout = QGridLayout(self.upgrades_box)
        self.upgrades_layout.setSpacing(6)
        self.upgrades_layout.setContentsMargins(11, 11, 11, 11)
        self.upgrades_layout.setObjectName(u"upgrades_layout")

        self.pickup_box_layout.addWidget(self.upgrades_box)

        self.translators_box = QGroupBox(self.pickups_scroll_contents)
        self.translators_box.setObjectName(u"translators_box")
        self.translators_layout = QGridLayout(self.translators_box)
        self.translators_layout.setSpacing(6)
        self.translators_layout.setContentsMargins(11, 11, 11, 11)
        self.translators_layout.setObjectName(u"translators_layout")

        self.pickup_box_layout.addWidget(self.translators_box)

        self.expansions_box = QGroupBox(self.pickups_scroll_contents)
        self.expansions_box.setObjectName(u"expansions_box")
        self.expansions_layout = QGridLayout(self.expansions_box)
        self.expansions_layout.setSpacing(6)
        self.expansions_layout.setContentsMargins(11, 11, 11, 11)
        self.expansions_layout.setObjectName(u"expansions_layout")

        self.pickup_box_layout.addWidget(self.expansions_box)

        self.keys_box = QGroupBox(self.pickups_scroll_contents)
        self.keys_box.setObjectName(u"keys_box")
        self.keys_layout = QGridLayout(self.keys_box)
        self.keys_layout.setSpacing(6)
        self.keys_layout.setContentsMargins(11, 11, 11, 11)
        self.keys_layout.setObjectName(u"keys_layout")

        self.pickup_box_layout.addWidget(self.keys_box)

        self.events_box = QGroupBox(self.pickups_scroll_contents)
        self.events_box.setObjectName(u"events_box")
        self.events_layout = QVBoxLayout(self.events_box)
        self.events_layout.setSpacing(6)
        self.events_layout.setContentsMargins(11, 11, 11, 11)
        self.events_layout.setObjectName(u"events_layout")

        self.pickup_box_layout.addWidget(self.events_box)

        self.pickups_scroll_area.setWidget(self.pickups_scroll_contents)

        self.tab_inventory_layout.addWidget(self.pickups_scroll_area)

        self.tab_widget.addTab(self.tab_inventory, "")
        self.tab_elevators = QWidget()
        self.tab_elevators.setObjectName(u"tab_elevators")
        self.tab_elevators_layout = QVBoxLayout(self.tab_elevators)
        self.tab_elevators_layout.setSpacing(6)
        self.tab_elevators_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_elevators_layout.setObjectName(u"tab_elevators_layout")
        self.tab_elevators_layout.setContentsMargins(0, 0, 0, 0)
        self.elevators_scroll_area = QScrollArea(self.tab_elevators)
        self.elevators_scroll_area.setObjectName(u"elevators_scroll_area")
        sizePolicy.setHeightForWidth(self.elevators_scroll_area.sizePolicy().hasHeightForWidth())
        self.elevators_scroll_area.setSizePolicy(sizePolicy)
        self.elevators_scroll_area.setMinimumSize(QSize(285, 0))
        self.elevators_scroll_area.setFrameShape(QFrame.NoFrame)
        self.elevators_scroll_area.setFrameShadow(QFrame.Plain)
        self.elevators_scroll_area.setWidgetResizable(True)
        self.elevators_scroll_contents = QWidget()
        self.elevators_scroll_contents.setObjectName(u"elevators_scroll_contents")
        self.elevators_scroll_contents.setGeometry(QRect(0, 0, 285, 409))
        self.elevators_scroll_layout = QVBoxLayout(self.elevators_scroll_contents)
        self.elevators_scroll_layout.setSpacing(6)
        self.elevators_scroll_layout.setContentsMargins(11, 11, 11, 11)
        self.elevators_scroll_layout.setObjectName(u"elevators_scroll_layout")
        self.elevators_scroll_layout.setContentsMargins(0, 0, 0, 0)
        self.elevators_scroll_area.setWidget(self.elevators_scroll_contents)

        self.tab_elevators_layout.addWidget(self.elevators_scroll_area)

        self.tab_widget.addTab(self.tab_elevators, "")
        self.tab_translator_gate = QWidget()
        self.tab_translator_gate.setObjectName(u"tab_translator_gate")
        self.tab_translator_gate_layout = QVBoxLayout(self.tab_translator_gate)
        self.tab_translator_gate_layout.setSpacing(6)
        self.tab_translator_gate_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_translator_gate_layout.setObjectName(u"tab_translator_gate_layout")
        self.tab_translator_gate_layout.setContentsMargins(0, 0, 0, 0)
        self.translator_gate_scroll_area = QScrollArea(self.tab_translator_gate)
        self.translator_gate_scroll_area.setObjectName(u"translator_gate_scroll_area")
        self.translator_gate_scroll_area.setWidgetResizable(True)
        self.translator_gate_scroll_contents = QWidget()
        self.translator_gate_scroll_contents.setObjectName(u"translator_gate_scroll_contents")
        self.translator_gate_scroll_contents.setGeometry(QRect(0, 0, 284, 407))
        self.translator_gate_scroll_layout = QGridLayout(self.translator_gate_scroll_contents)
        self.translator_gate_scroll_layout.setSpacing(6)
        self.translator_gate_scroll_layout.setContentsMargins(11, 11, 11, 11)
        self.translator_gate_scroll_layout.setObjectName(u"translator_gate_scroll_layout")
        self.translator_gate_scroll_layout.setContentsMargins(3, 3, 3, 3)
        self.translator_gate_scroll_area.setWidget(self.translator_gate_scroll_contents)

        self.tab_translator_gate_layout.addWidget(self.translator_gate_scroll_area)

        self.tab_widget.addTab(self.tab_translator_gate, "")

        self.gridLayout.addWidget(self.tab_widget, 2, 0, 1, 1)

        self.map_tab_widget = QTabWidget(self.centralWidget)
        self.map_tab_widget.setObjectName(u"map_tab_widget")
        self.tab_text_map = QWidget()
        self.tab_text_map.setObjectName(u"tab_text_map")
        self.tab_text_map_layout = QVBoxLayout(self.tab_text_map)
        self.tab_text_map_layout.setSpacing(6)
        self.tab_text_map_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_text_map_layout.setObjectName(u"tab_text_map_layout")
        self.tab_text_map_layout.setContentsMargins(4, 4, 4, 4)
        self.resource_filter_check = QCheckBox(self.tab_text_map)
        self.resource_filter_check.setObjectName(u"resource_filter_check")
        self.resource_filter_check.setChecked(True)

        self.tab_text_map_layout.addWidget(self.resource_filter_check)

        self.hide_collected_resources_check = QCheckBox(self.tab_text_map)
        self.hide_collected_resources_check.setObjectName(u"hide_collected_resources_check")

        self.tab_text_map_layout.addWidget(self.hide_collected_resources_check)

        self.current_location_label = QLabel(self.tab_text_map)
        self.current_location_label.setObjectName(u"current_location_label")

        self.tab_text_map_layout.addWidget(self.current_location_label)

        self.possible_locations_tree = QTreeWidget(self.tab_text_map)
        self.possible_locations_tree.setObjectName(u"possible_locations_tree")

        self.tab_text_map_layout.addWidget(self.possible_locations_tree)

        self.map_tab_widget.addTab(self.tab_text_map, "")
        self.tab_graph_map = QWidget()
        self.tab_graph_map.setObjectName(u"tab_graph_map")
        self.tab_graph_map_layout = QVBoxLayout(self.tab_graph_map)
        self.tab_graph_map_layout.setSpacing(6)
        self.tab_graph_map_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_graph_map_layout.setObjectName(u"tab_graph_map_layout")
        self.tab_graph_map_layout.setContentsMargins(0, 0, 0, 0)
        self.graph_map_world_combo = QComboBox(self.tab_graph_map)
        self.graph_map_world_combo.setObjectName(u"graph_map_world_combo")

        self.tab_graph_map_layout.addWidget(self.graph_map_world_combo)

        self.map_tab_widget.addTab(self.tab_graph_map, "")
        self.tab_actions = QWidget()
        self.tab_actions.setObjectName(u"tab_actions")
        self.tab_actions_layout = QVBoxLayout(self.tab_actions)
        self.tab_actions_layout.setSpacing(6)
        self.tab_actions_layout.setContentsMargins(11, 11, 11, 11)
        self.tab_actions_layout.setObjectName(u"tab_actions_layout")
        self.tab_actions_layout.setContentsMargins(4, 4, 4, 4)
        self.undo_last_action_button = QPushButton(self.tab_actions)
        self.undo_last_action_button.setObjectName(u"undo_last_action_button")

        self.tab_actions_layout.addWidget(self.undo_last_action_button)

        self.actions_title_label = QLabel(self.tab_actions)
        self.actions_title_label.setObjectName(u"actions_title_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.actions_title_label.sizePolicy().hasHeightForWidth())
        self.actions_title_label.setSizePolicy(sizePolicy1)
        self.actions_title_label.setWordWrap(True)

        self.tab_actions_layout.addWidget(self.actions_title_label)

        self.actions_list = QListWidget(self.tab_actions)
        self.actions_list.setObjectName(u"actions_list")

        self.tab_actions_layout.addWidget(self.actions_list)

        self.map_tab_widget.addTab(self.tab_actions, "")

        self.gridLayout.addWidget(self.map_tab_widget, 2, 1, 1, 2)

        TrackerWindow.setCentralWidget(self.centralWidget)
        self.menu_bar = QMenuBar(TrackerWindow)
        self.menu_bar.setObjectName(u"menu_bar")
        self.menu_bar.setGeometry(QRect(0, 0, 1040, 21))
        self.menu_tracker = QMenu(self.menu_bar)
        self.menu_tracker.setObjectName(u"menu_tracker")
        TrackerWindow.setMenuBar(self.menu_bar)

        self.menu_bar.addAction(self.menu_tracker.menuAction())
        self.menu_tracker.addAction(self.menu_reset_action)

        self.retranslateUi(TrackerWindow)

        self.tab_widget.setCurrentIndex(0)
        self.map_tab_widget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(TrackerWindow)
    # setupUi

    def retranslateUi(self, TrackerWindow):
        TrackerWindow.setWindowTitle(QCoreApplication.translate("TrackerWindow", u"Tracker", None))
        self.actionAction.setText(QCoreApplication.translate("TrackerWindow", u"Action", None))
        self.menu_reset_action.setText(QCoreApplication.translate("TrackerWindow", u"Reset", None))
        self.configuration_label.setText(QCoreApplication.translate("TrackerWindow", u"Trick Level: ???; Elevators: Vanilla; Item Loss: ???", None))
        self.upgrades_box.setTitle(QCoreApplication.translate("TrackerWindow", u"Upgrades", None))
        self.translators_box.setTitle(QCoreApplication.translate("TrackerWindow", u"Translators", None))
        self.expansions_box.setTitle(QCoreApplication.translate("TrackerWindow", u"Expansions", None))
        self.keys_box.setTitle(QCoreApplication.translate("TrackerWindow", u"Keys", None))
        self.events_box.setTitle(QCoreApplication.translate("TrackerWindow", u"Events", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_inventory), QCoreApplication.translate("TrackerWindow", u"Inventory", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_elevators), QCoreApplication.translate("TrackerWindow", u"Elevators", None))
        self.tab_widget.setTabText(self.tab_widget.indexOf(self.tab_translator_gate), QCoreApplication.translate("TrackerWindow", u"Translator Gate", None))
        self.resource_filter_check.setText(QCoreApplication.translate("TrackerWindow", u"Show only resources", None))
        self.hide_collected_resources_check.setText(QCoreApplication.translate("TrackerWindow", u"Hide collected resources", None))
        self.current_location_label.setText(QCoreApplication.translate("TrackerWindow", u"Current location:", None))
        ___qtreewidgetitem = self.possible_locations_tree.headerItem()
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("TrackerWindow", u"Accessible Locations", None));
        self.map_tab_widget.setTabText(self.map_tab_widget.indexOf(self.tab_text_map), QCoreApplication.translate("TrackerWindow", u"Text Map", None))
        self.map_tab_widget.setTabText(self.map_tab_widget.indexOf(self.tab_graph_map), QCoreApplication.translate("TrackerWindow", u"Graph Map", None))
        self.undo_last_action_button.setText(QCoreApplication.translate("TrackerWindow", u"Undo last action", None))
        self.actions_title_label.setText(QCoreApplication.translate("TrackerWindow", u"<html><head/><body><p>History of all actions that have been performed.</p><p>Press &quot;Undo last action&quot; to remove the last action from the list.</p></body></html>", None))
        self.map_tab_widget.setTabText(self.map_tab_widget.indexOf(self.tab_actions), QCoreApplication.translate("TrackerWindow", u"Actions", None))
        self.menu_tracker.setTitle(QCoreApplication.translate("TrackerWindow", u"Tracker", None))
    # retranslateUi




import collections
import copy
import dataclasses

import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description import data_reader
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.hint import Hint
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import PickupNode
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import ConditionalResources, ResourceConversion, PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.games.game import RandovaniaGame
from randovania.generator import generator
from randovania.generator.item_pool import pickup_creator, pool_creator
from randovania.layout import game_patches_serializer
from randovania.network_common.pickup_serializer import BitPackPickupEntry
from randovania.layout.major_item_state import MajorItemState
from randovania.layout.permalink import Permalink
from randovania.layout.trick_level import LayoutTrickLevel, TrickLevelConfiguration


@pytest.fixture(
    params=[
        {},
        {"starting_item": "Morph Ball"},
        {"elevator": [1572998, "Temple Grounds/Transport to Agon Wastes"]},
        {"translator": [(10, "Mining Plaza", "Cobalt Translator"), (12, "Great Bridge", "Emerald Translator")]},
        {"pickup": "Morph Ball Bomb"},
        {"hint": [1000, {"hint_type": "location",
                         "dark_temple": None,
                         "precision": {"location": "detailed", "item": "detailed", "relative": None,
                                       "include_owner": False},
                         "target": 50}]},
    ],
    name="patches_with_data")
def _patches_with_data(request, echoes_game_data, echoes_item_database):
    game = data_reader.decode_data(echoes_game_data)

    data = {
        "starting_location": "Temple Grounds/Landing Site",
        "starting_items": {},
        "elevators": {
            "Temple Grounds/Temple Transport C": "Great Temple/Temple Transport C",
            "Temple Grounds/Transport to Agon Wastes": "Agon Wastes/Transport to Temple Grounds",
            "Temple Grounds/Transport to Torvus Bog": "Torvus Bog/Transport to Temple Grounds",
            "Temple Grounds/Temple Transport B": "Great Temple/Temple Transport B",
            "Temple Grounds/Transport to Sanctuary Fortress": "Sanctuary Fortress/Transport to Temple Grounds",
            "Temple Grounds/Temple Transport A": "Great Temple/Temple Transport A",
            "Great Temple/Temple Transport A": "Temple Grounds/Temple Transport A",
            "Great Temple/Temple Transport C": "Temple Grounds/Temple Transport C",
            "Great Temple/Temple Transport B": "Temple Grounds/Temple Transport B",
            "Sky Temple Grounds/Sky Temple Gateway": "Sky Temple/Sky Temple Energy Controller",
            "Sky Temple/Sky Temple Energy Controller": "Sky Temple Grounds/Sky Temple Gateway",
            "Agon Wastes/Transport to Temple Grounds": "Temple Grounds/Transport to Agon Wastes",
            "Agon Wastes/Transport to Torvus Bog": "Torvus Bog/Transport to Agon Wastes",
            "Agon Wastes/Transport to Sanctuary Fortress": "Sanctuary Fortress/Transport to Agon Wastes",
            "Torvus Bog/Transport to Temple Grounds": "Temple Grounds/Transport to Torvus Bog",
            "Torvus Bog/Transport to Agon Wastes": "Agon Wastes/Transport to Torvus Bog",
            "Torvus Bog/Transport to Sanctuary Fortress": "Sanctuary Fortress/Transport to Torvus Bog",
            "Sanctuary Fortress/Transport to Temple Grounds": "Temple Grounds/Transport to Sanctuary Fortress",
            "Sanctuary Fortress/Transport to Agon Wastes": "Agon Wastes/Transport to Sanctuary Fortress",
            "Sanctuary Fortress/Transport to Torvus Bog": "Torvus Bog/Transport to Sanctuary Fortress",
            "Sanctuary Fortress/Aerie": "Sanctuary Fortress/Aerie Transport Station",
            "Sanctuary Fortress/Aerie Transport Station": "Sanctuary Fortress/Aerie",
        },
        "translators": {},
        "locations": {},
        "hints": {}
    }
    patches = dataclasses.replace(game.create_game_patches(), player_index=0)

    locations = collections.defaultdict(dict)
    for world, area, node in game.world_list.all_worlds_areas_nodes:
        if node.is_resource_node and isinstance(node, PickupNode):
            world_name = world.dark_name if area.in_dark_aether else world.name
            locations[world_name][game.world_list.node_name(node)] = game_patches_serializer._ETM_NAME

    data["locations"] = {
        world: {
            area: item
            for area, item in sorted(locations[world].items())
        }
        for world in sorted(locations.keys())
    }

    if request.param.get("starting_item"):
        item_name = request.param.get("starting_item")
        patches = patches.assign_extra_initial_items({
            find_resource_info_with_long_name(game.resource_database.item, item_name): 1,
        })
        data["starting_items"][item_name] = 1

    if request.param.get("elevator"):
        elevator_id, elevator_source = request.param.get("elevator")
        elevator_connection = copy.copy(patches.elevator_connection)
        elevator_connection[elevator_id] = game.starting_location

        patches = dataclasses.replace(patches, elevator_connection=elevator_connection)
        data["elevators"][elevator_source] = "Temple Grounds/Landing Site"

    if request.param.get("translator"):
        gates = {}
        for index, gate_name, translator in request.param.get("translator"):
            gates[TranslatorGate(index)] = find_resource_info_with_long_name(game.resource_database.item, translator)
            data["translators"][gate_name] = translator

        patches = patches.assign_gate_assignment(gates)

    if request.param.get("pickup"):
        pickup_name = request.param.get("pickup")
        pickup = pickup_creator.create_major_item(echoes_item_database.major_items[pickup_name],
                                                  MajorItemState(), True, game.resource_database,
                                                  None, False)

        patches = patches.assign_new_pickups([(PickupIndex(5), PickupTarget(pickup, 0))])
        data["locations"]["Temple Grounds"]['Transport to Agon Wastes/Pickup (Missile)'] = pickup_name

    if request.param.get("hint"):
        asset, hint = request.param.get("hint")
        patches = patches.assign_hint(LogbookAsset(asset), Hint.from_json(hint))
        data["hints"][str(asset)] = hint

    return data, patches


def test_encode(patches_with_data, echoes_game_data):
    expected, patches = patches_with_data

    # Run
    encoded = game_patches_serializer.serialize_single(0, 1, patches, echoes_game_data)

    # Assert
    for key, value in expected["locations"].items():
        assert encoded["locations"][key] == value
    assert encoded == expected


def test_decode(patches_with_data, default_layout_configuration):
    encoded, expected = patches_with_data

    game = data_reader.decode_data(default_layout_configuration.game_data)
    pool = pool_creator.calculate_pool_results(default_layout_configuration, game.resource_database)

    # Run
    decoded = game_patches_serializer.decode_single(0, {0: pool}, game, encoded, default_layout_configuration)

    # Assert
    assert decoded == expected


@pytest.mark.parametrize("has_convert", [False, True])
def test_bit_pack_pickup_entry(has_convert: bool, echoes_resource_database):
    # Setup
    name = "Some Random Name"
    if has_convert:
        convert_resources = (
            ResourceConversion(
                find_resource_info_with_long_name(echoes_resource_database.item, "Morph Ball"),
                find_resource_info_with_long_name(echoes_resource_database.item, "Item Percentage")
            ),
        )
    else:
        convert_resources = ()

    pickup = PickupEntry(
        name=name,
        model_index=26,
        item_category=ItemCategory.TEMPLE_KEY,
        broad_category=ItemCategory.KEY,
        resources=(
            ConditionalResources(
                "Morph Ball", None,
                (
                    (find_resource_info_with_long_name(echoes_resource_database.item, "Morph Ball"), 2),
                    (find_resource_info_with_long_name(echoes_resource_database.item, "Item Percentage"), 5),
                ),
            ),
            ConditionalResources(
                "Grapple Beam", find_resource_info_with_long_name(echoes_resource_database.item, "Morph Ball"),
                (
                    (find_resource_info_with_long_name(echoes_resource_database.item, "Grapple Beam"), 3),
                ),
            )
        ),
        convert_resources=convert_resources
    )

    # Run
    encoded = bitpacking.pack_value(BitPackPickupEntry(pickup, echoes_resource_database))
    decoder = BitPackDecoder(encoded)
    decoded = BitPackPickupEntry.bit_pack_unpack(decoder, echoes_resource_database)

    # Assert
    assert pickup == decoded


def test_round_trip_generated_patches(echoes_game_data, default_preset):
    # Setup
    preset = dataclasses.replace(
        default_preset,
        base_preset_name=default_preset.name,
        configuration=dataclasses.replace(
            default_preset.configuration,
            trick_level=TrickLevelConfiguration(
                minimal_logic=True,
                specific_levels={},
                game=RandovaniaGame.PRIME2,
            )
        )
    )

    all_patches = generator._async_create_description(
        permalink=Permalink(
            seed_number=1000,
            spoiler=True,
            presets={0: preset},
        ),
        status_update=lambda x: None,
        attempts=0,
    ).all_patches

    # Run
    encoded = game_patches_serializer.serialize(all_patches, {0: echoes_game_data})
    decoded = game_patches_serializer.decode(encoded, {0: preset.configuration})

    # Assert
    assert all_patches == decoded



import dataclasses
from typing import Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder


@dataclasses.dataclass(frozen=True)
class AmmoState(BitPackValue):
    variance: int = 0
    pickup_count: int = 0
    requires_major_item: bool = True

    @classmethod
    def maximum_pickup_count(cls) -> int:
        return 64

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield self.pickup_count, AmmoState.maximum_pickup_count()
        yield from bitpacking.encode_bool(self.requires_major_item)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata) -> "AmmoState":
        pickup_count = decoder.decode_single(cls.maximum_pickup_count())
        requires_major_item = bitpacking.decode_bool(decoder)

        return cls(
            variance=0,
            pickup_count=pickup_count,
            requires_major_item=requires_major_item,
        )

    @property
    def as_json(self) -> dict:
        result = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            result[field.name] = value

        return result

    @classmethod
    def from_json(cls, value: dict) -> "AmmoState":
        kwargs = {}

        for field in dataclasses.fields(cls):
            if field.name in value:
                kwargs[field.name] = value[field.name]

        return cls(**kwargs)



import dataclasses
from enum import Enum
from typing import Optional, NamedTuple, Tuple

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.dock import DockWeakness, DockConnection
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.requirements import ResourceRequirement, Requirement, RequirementAnd
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import ResourceInfo, ResourceGain, CurrentResources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.translator_gate import TranslatorGate


class NodeLocation(NamedTuple):
    x: float
    y: float
    z: float


@dataclasses.dataclass(frozen=True)
class Node:
    name: str
    heal: bool
    location: Optional[NodeLocation]
    index: int

    def __lt__(self, other):
        return self.name < other.name

    def __hash__(self):
        return hash((self.index, self.name))

    @property
    def is_resource_node(self) -> bool:
        return False

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        return Requirement.trivial()


@dataclasses.dataclass(frozen=True)
class ResourceNode(Node):
    @property
    def is_resource_node(self) -> bool:
        return True

    def resource(self) -> ResourceInfo:
        raise NotImplementedError

    def can_collect(self, patches: GamePatches, current_resources: CurrentResources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        raise NotImplementedError

    def resource_gain_on_collect(self, patches: GamePatches, current_resources: CurrentResources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class GenericNode(Node):
    pass


@dataclasses.dataclass(frozen=True)
class DockNode(Node):
    dock_index: int
    default_connection: DockConnection
    default_dock_weakness: DockWeakness

    def __hash__(self):
        return hash((self.index, self.name, self.dock_index))

    def __repr__(self):
        return "DockNode({!r}/{} -> {})".format(self.name, self.dock_index, self.default_connection)


@dataclasses.dataclass(frozen=True)
class TeleporterNode(Node):
    teleporter_instance_id: Optional[int]
    default_connection: AreaLocation
    scan_asset_id: Optional[int]
    keep_name_when_vanilla: bool
    editable: bool

    def __post_init__(self):
        if self.editable and self.teleporter_instance_id is None:
            raise ValueError(f"{self!r} is editable, but teleporter_instance_id is None")

    def __repr__(self):
        return "TeleporterNode({!r} -> {})".format(self.name, self.default_connection)


@dataclasses.dataclass(frozen=True)
class PickupNode(ResourceNode):
    pickup_index: PickupIndex
    major_location: bool

    def __repr__(self):
        return "PickupNode({!r} -> {})".format(self.name, self.pickup_index.index)

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        # FIXME: using non-resource as key in CurrentResources
        if current_resources.get("add_self_as_requirement_to_resources") == 1:
            return ResourceRequirement(self.pickup_index, 1, False)
        else:
            return Requirement.trivial()

    def resource(self) -> ResourceInfo:
        return self.pickup_index

    def can_collect(self, patches: GamePatches, current_resources: CurrentResources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        return current_resources.get(self.pickup_index, 0) == 0

    def resource_gain_on_collect(self, patches: GamePatches, current_resources: CurrentResources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        yield self.pickup_index, 1

        target = patches.pickup_assignment.get(self.pickup_index)
        if target is not None and target.player == patches.player_index:
            yield from target.pickup.resource_gain(current_resources)


@dataclasses.dataclass(frozen=True)
class EventNode(ResourceNode):
    event: ResourceInfo

    def __repr__(self):
        return "EventNode({!r} -> {})".format(self.name, self.event.long_name)

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        if current_resources.get("add_self_as_requirement_to_resources") == 1:
            return ResourceRequirement(self.event, 1, False)
        else:
            return Requirement.trivial()

    def resource(self) -> ResourceInfo:
        return self.event

    def can_collect(self, patches: GamePatches, current_resources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        return current_resources.get(self.event, 0) == 0

    def resource_gain_on_collect(self, patches: GamePatches, current_resources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        yield self.event, 1


@dataclasses.dataclass(frozen=True)
class TranslatorGateNode(ResourceNode):
    gate: TranslatorGate
    scan_visor: ItemResourceInfo

    def __repr__(self):
        return "TranslatorGateNode({!r} -> {})".format(self.name, self.gate.index)

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        return RequirementAnd([
            ResourceRequirement(patches.translator_gates[self.gate], 1, False),
            ResourceRequirement(self.scan_visor, 1, False),
        ])

    def resource(self) -> ResourceInfo:
        return self.gate

    def can_collect(self, patches: GamePatches, current_resources: CurrentResources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        """
        Checks if this TranslatorGate can be opened with the given resources and translator gate mapping
        :param patches:
        :param current_resources:
        :param all_nodes:
        :return:
        """
        if current_resources.get(self.gate, 0) != 0:
            return False
        translator = patches.translator_gates[self.gate]
        return current_resources.get(self.scan_visor, 0) > 0 and current_resources.get(translator, 0) > 0

    def resource_gain_on_collect(self, patches: GamePatches, current_resources: CurrentResources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        yield self.gate, 1


class LoreType(Enum):
    LUMINOTH_LORE = "luminoth-lore"
    LUMINOTH_WARRIOR = "luminoth-warrior"
    PIRATE_LORE = "pirate-lore"
    SKY_TEMPLE_KEY_HINT = "sky-temple-key-hint"

    @property
    def holds_generic_hint(self) -> bool:
        return self in {LoreType.LUMINOTH_LORE, LoreType.PIRATE_LORE}

    @property
    def long_name(self) -> str:
        return _LORE_TYPE_LONG_NAME[self]


_LORE_TYPE_LONG_NAME = {
    LoreType.LUMINOTH_LORE: "Luminoth Lore",
    LoreType.LUMINOTH_WARRIOR: "Keybearer Corpse",
    LoreType.PIRATE_LORE: "Pirate Lore",
    LoreType.SKY_TEMPLE_KEY_HINT: "Sky Temple Key Hint",
}


@dataclasses.dataclass(frozen=True)
class LogbookNode(ResourceNode):
    string_asset_id: int
    scan_visor: ItemResourceInfo
    lore_type: LoreType
    required_translator: Optional[ItemResourceInfo]
    hint_index: Optional[int]

    def __repr__(self):
        extra = None
        if self.required_translator is not None:
            extra = self.required_translator.short_name
        elif self.hint_index is not None:
            extra = self.hint_index
        return "LogbookNode({!r} -> {}/{}{})".format(
            self.name,
            self.string_asset_id,
            self.lore_type.value,
            f"/{extra}" if extra is not None else ""
        )

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        items = [ResourceRequirement(self.scan_visor, 1, False)]
        if self.required_translator is not None:
            items.append(ResourceRequirement(self.required_translator, 1, False))

        return RequirementAnd(items)

    def resource(self) -> ResourceInfo:
        return LogbookAsset(self.string_asset_id)

    def can_collect(self, patches: GamePatches, current_resources: CurrentResources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        """
        Checks if this TranslatorGate can be opened with the given resources and translator gate mapping
        :param patches:
        :param current_resources:
        :param all_nodes:
        :return:
        """
        if current_resources.get(self.resource(), 0) != 0:
            return False

        if current_resources.get(self.scan_visor, 0) == 0:
            return False

        if self.required_translator is not None:
            return current_resources.get(self.required_translator, 0) > 0
        else:
            return True

    def resource_gain_on_collect(self, patches: GamePatches, current_resources: CurrentResources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        yield self.resource(), 1


@dataclasses.dataclass(frozen=True)
class PlayerShipNode(ResourceNode):
    is_unlocked: Requirement
    item_to_summon: ItemResourceInfo

    @property
    def visor_requirement(self):
        return ResourceRequirement(self.item_to_summon, 1, False)

    def requirement_to_leave(self, patches: GamePatches, current_resources: CurrentResources) -> Requirement:
        return RequirementAnd([self.is_unlocked, ResourceRequirement(self.resource(), 1, False)])

    def resource(self) -> SimpleResourceInfo:
        return SimpleResourceInfo(self.index, f"Ship Node {self.index}", f"Ship{self.index}", ResourceType.SHIP_NODE)

    def can_collect(self, patches: GamePatches, current_resources: CurrentResources,
                    all_nodes: Tuple[Node, ...]) -> bool:
        """
        Checks if this TranslatorGate can be opened with the given resources and translator gate mapping
        :param patches:
        :param current_resources:
        :param all_nodes:
        :return:
        """
        if current_resources.get(self.item_to_summon, 0) == 0 and current_resources.get(self.resource(), 0) == 0:
            return False

        return any(
            current_resources.get(node.resource(), 0) == 0
            for node in all_nodes
            if isinstance(node, PlayerShipNode) and node.is_unlocked.satisfied(current_resources, 0)
        )

    def resource_gain_on_collect(self, patches: GamePatches, current_resources: CurrentResources,
                                 all_nodes: Tuple[Node, ...]) -> ResourceGain:
        for node in all_nodes:
            if isinstance(node, PlayerShipNode) and node.is_unlocked.satisfied(current_resources, 0):
                yield node.resource(), 1



from pathlib import Path
from typing import Optional

from PySide2.QtWidgets import QMessageBox, QDialog

from randovania.gui.generated.game_input_dialog_ui import Ui_GameInputDialog
from randovania.gui.lib import common_qt_lib
from randovania.interface_common import simplified_patcher, game_workdir
from randovania.interface_common.options import Options

_VALID_GAME_TEXT = "(internal Metroid Prime 2 copy)"


class GameInputDialog(QDialog, Ui_GameInputDialog):
    _options: Options
    _has_game: bool
    _current_lock_state: bool = True

    def __init__(self, options: Options, default_iso_name: str, spoiler: bool):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self._options = options
        self.default_iso_name = default_iso_name
        self.check_extracted_game()

        # Input
        self.input_file_edit.textChanged.connect(self._validate_input_file)
        self.input_file_button.clicked.connect(self._on_input_file_button)

        # Output
        self.output_file_edit.textChanged.connect(self._validate_output_file)
        self.output_file_button.clicked.connect(self._on_output_file_button)

        # Spoiler
        self.auto_save_spoiler_check.setEnabled(spoiler)
        self.auto_save_spoiler_check.setChecked(options.auto_save_spoiler)

        # Accept/Reject
        self.accept_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

        self.input_file_edit.has_error = False
        self.output_file_edit.has_error = False

        if options.output_directory is not None:
            self.output_file_edit.setText(str(options.output_directory.joinpath(self.default_iso_name)))

        self._validate_input_file()
        self._validate_output_file()

    # Getters
    @property
    def input_file(self) -> Optional[Path]:
        if self._has_game:
            return None
        else:
            return Path(self.input_file_edit.text())

    @property
    def output_file(self) -> Path:
        return Path(self.output_file_edit.text())

    @property
    def auto_save_spoiler(self) -> bool:
        return self.auto_save_spoiler_check.isChecked()

    def _update_accept_button(self):
        self.accept_button.setEnabled(not (self.input_file_edit.has_error or self.output_file_edit.has_error))

    # Checks
    def check_extracted_game(self):
        result = game_workdir.discover_game(self._options.game_files_path)
        self._has_game = result is not None
        self.input_file_edit.setEnabled(not self._has_game)

        if not self._has_game:
            self.input_file_button.setText("Select File")
        else:
            self.input_file_button.setText("Delete internal copy")
            game_id, _ = result
            if game_id.startswith("G2M"):
                self.input_file_edit.setText(_VALID_GAME_TEXT)
            else:
                self.input_file_edit.setText("(internal unknown game copy)")

    # Input file
    def _validate_input_file(self):
        if self._has_game:
            has_error = self.input_file_edit.text() != _VALID_GAME_TEXT
        else:
            has_error = not self.input_file.is_file()

        common_qt_lib.set_error_border_stylesheet(self.input_file_edit, has_error)
        self._update_accept_button()

    def _on_input_file_button(self):
        if self._has_game:
            simplified_patcher.delete_files_location(self._options)
            self.input_file_edit.setText("")
            self.check_extracted_game()
        else:
            input_file = common_qt_lib.prompt_user_for_input_iso(self)
            if input_file is not None:
                self.input_file_edit.setText(str(input_file.absolute()))

    # Output File
    def _validate_output_file(self):
        output_file = self.output_file
        has_error = output_file.is_dir() or not output_file.parent.is_dir()

        common_qt_lib.set_error_border_stylesheet(self.output_file_edit, has_error)
        self._update_accept_button()

    def _on_output_file_button(self):
        output_file = common_qt_lib.prompt_user_for_output_iso(self, self.default_iso_name)
        if output_file is None:
            return

        output_file = output_file.absolute()

        try:
            output_file.parent.mkdir(parents=True, exist_ok=True)
        except (FileNotFoundError, OSError) as error:
            QMessageBox.warning(self,
                                "Invalid output file",
                                "Unable to use '{}' as output file: {}".format(output_file, error),
                                )
            return

        self.output_file_edit.setText(str(output_file))



import collections
import dataclasses
import functools
import re
from typing import Dict, Optional, List

from PySide2 import QtCore, QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QComboBox, QDialog, QGroupBox, QVBoxLayout

from randovania.game_description import default_database
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.node import PickupNode
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.gui.dialog.trick_details_popup import TrickDetailsPopup
from randovania.gui.generated.logic_settings_window_ui import Ui_LogicSettingsWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.common_qt_lib import set_combo_with_value
from randovania.gui.lib.trick_lib import difficulties_for_trick, used_tricks
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.preset_settings.echoes_beam_configuration_tab import PresetEchoesBeamConfiguration
from randovania.gui.preset_settings.echoes_goal_tab import PresetEchoesGoal
from randovania.gui.preset_settings.echoes_hints_tab import PresetEchoesHints
from randovania.gui.preset_settings.echoes_patches_tab import PresetEchoesPatches
from randovania.gui.preset_settings.echoes_translators_tab import PresetEchoesTranslators
from randovania.gui.preset_settings.item_pool_tab import PresetItemPool
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.enum_lib import iterate_enum
from randovania.interface_common.options import Options
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.available_locations import RandomizationMode
from randovania.layout.damage_strictness import LayoutDamageStrictness
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.preset import Preset
from randovania.layout.starting_location import StartingLocation
from randovania.layout.trick_level import LayoutTrickLevel


def _update_options_when_true(options: Options, field_name: str, new_value, checked: bool):
    if checked:
        with options:
            setattr(options, field_name, new_value)


def dark_world_flags(world: World):
    yield False
    if world.dark_name is not None:
        yield True


def _update_options_by_value(options: Options, combo: QComboBox, new_index: int):
    with options:
        setattr(options, combo.options_field_name, combo.currentData())


class LogicSettingsWindow(QDialog, Ui_LogicSettingsWindow):
    _extra_tabs: List[PresetTab]
    _combo_for_gate: Dict[TranslatorGate, QComboBox]
    _location_pool_for_node: Dict[PickupNode, QtWidgets.QCheckBox]
    _starting_location_for_area: Dict[int, QtWidgets.QCheckBox]
    _starting_location_for_world: Dict[str, QtWidgets.QCheckBox]
    _slider_for_trick: Dict[TrickResourceInfo, QtWidgets.QSlider]
    _editor: PresetEditor
    world_list: WorldList
    _during_batch_check_update: bool = False

    def __init__(self, window_manager: Optional[WindowManager], editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self._editor = editor
        self._window_manager = window_manager
        self._extra_tabs = []

        self.game_enum = editor.game
        self.game_description = default_database.game_description_for(self.game_enum)
        self.world_list = self.game_description.world_list
        self.resource_database = self.game_description.resource_database

        if self.game_enum == RandovaniaGame.PRIME2:
            self._extra_tabs.append(PresetEchoesGoal(editor))
            self._extra_tabs.append(PresetEchoesHints(editor))
            self._extra_tabs.append(PresetEchoesTranslators(editor))
            self._extra_tabs.append(PresetEchoesBeamConfiguration(editor))
            self._extra_tabs.append(PresetEchoesPatches(editor))

        elif self.game_enum == RandovaniaGame.PRIME3:
            pass

        self._extra_tabs.append(PresetItemPool(editor))

        for extra_tab in self._extra_tabs:
            if extra_tab.uses_patches_tab:
                tab = self.patches_tab_widget
            else:
                tab = self.logic_tab_widget
            tab.addTab(extra_tab, extra_tab.tab_title)

        self.name_edit.textEdited.connect(self._edit_name)
        self.setup_trick_level_elements()
        self.setup_damage_elements()
        self.setup_elevator_elements()
        self.setup_starting_area_elements()
        self.setup_location_pool_elements()

        # Alignment
        self.trick_level_layout.setAlignment(QtCore.Qt.AlignTop)
        self.elevator_layout.setAlignment(QtCore.Qt.AlignTop)
        self.starting_area_layout.setAlignment(QtCore.Qt.AlignTop)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    # Options
    def on_preset_changed(self, preset: Preset):
        for extra_tab in self._extra_tabs:
            extra_tab.on_preset_changed(preset)

        # Variables
        config = preset.configuration

        # Title
        common_qt_lib.set_edit_if_different(self.name_edit, preset.name)

        # Trick Level
        trick_level_configuration = preset.configuration.trick_level
        self.trick_level_minimal_logic_check.setChecked(trick_level_configuration.minimal_logic)

        for trick, slider in self._slider_for_trick.items():
            assert self._slider_for_trick[trick] is slider
            slider.setValue(trick_level_configuration.level_for_trick(trick).as_number)
            slider.setEnabled(not trick_level_configuration.minimal_logic)

        # Damage
        set_combo_with_value(self.damage_strictness_combo, config.damage_strictness)
        self.energy_tank_capacity_spin_box.setValue(config.energy_per_tank)
        self.dangerous_tank_check.setChecked(config.dangerous_energy_tank)
        if self.game_enum == RandovaniaGame.PRIME2:
            self.safe_zone_logic_heal_check.setChecked(config.safe_zone.fully_heal)
            self.safe_zone_regen_spin.setValue(config.safe_zone.heal_per_second)
            self.varia_suit_spin_box.setValue(config.varia_suit_damage)
            self.dark_suit_spin_box.setValue(config.dark_suit_damage)

        # Elevator
        set_combo_with_value(self.elevators_combo, config.elevators)

        # Starting Area
        self.on_preset_changed_starting_area(preset)

        # Location Pool
        available_locations = config.available_locations
        set_combo_with_value(self.randomization_mode_combo, available_locations.randomization_mode)

        self._during_batch_check_update = True
        for node, check in self._location_pool_for_node.items():
            check.setChecked(node.pickup_index not in available_locations.excluded_indices)
            check.setEnabled(available_locations.randomization_mode == RandomizationMode.FULL or node.major_location)
        self._during_batch_check_update = False

    def _edit_name(self, value: str):
        with self._editor as editor:
            editor.name = value

    # Trick Level

    def _create_difficulty_details_row(self):
        row = 1

        trick_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
        trick_label.setWordWrap(True)
        trick_label.setText("Difficulty Details")

        self.trick_difficulties_layout.addWidget(trick_label, row, 1, 1, -1)

        slider_layout = QtWidgets.QGridLayout()
        slider_layout.setHorizontalSpacing(0)
        for i in range(12):
            slider_layout.setColumnStretch(i, 1)
        #
        # if self._window_manager is not None:
        #     for i, trick_level in enumerate(LayoutTrickLevel):
        #         if trick_level not in {LayoutTrickLevel.NO_TRICKS, LayoutTrickLevel.MINIMAL_LOGIC}:
        #             tool_button = QtWidgets.QToolButton(self.trick_level_scroll_contents)
        #             tool_button.setText(trick_level.long_name)
        #             tool_button.clicked.connect(functools.partial(self._open_difficulty_details_popup, trick_level))
        #
        #             slider_layout.addWidget(tool_button, 1, 2 * i, 1, 2)

        self.trick_difficulties_layout.addLayout(slider_layout, row, 2, 1, 1)

    def setup_trick_level_elements(self):
        self.trick_level_minimal_logic_check.stateChanged.connect(self._on_trick_level_minimal_logic_check)

        self.trick_difficulties_layout = QtWidgets.QGridLayout()
        self._slider_for_trick = {}

        tricks_in_use = used_tricks(self.game_description)

        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)

        self._create_difficulty_details_row()

        row = 2
        for trick in sorted(self.resource_database.trick, key=lambda _trick: _trick.long_name):
            if trick not in tricks_in_use:
                continue

            if row > 1:
                self.trick_difficulties_layout.addItem(QtWidgets.QSpacerItem(20, 40,
                                                                             QtWidgets.QSizePolicy.Minimum,
                                                                             QtWidgets.QSizePolicy.Expanding))

            trick_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
            trick_label.setSizePolicy(size_policy)
            trick_label.setWordWrap(True)
            trick_label.setFixedWidth(100)
            trick_label.setText(trick.long_name)

            self.trick_difficulties_layout.addWidget(trick_label, row, 1, 1, 1)

            slider_layout = QtWidgets.QGridLayout()
            slider_layout.setHorizontalSpacing(0)
            for i in range(12):
                slider_layout.setColumnStretch(i, 1)

            horizontal_slider = QtWidgets.QSlider(self.trick_level_scroll_contents)
            horizontal_slider.setMaximum(5)
            horizontal_slider.setPageStep(2)
            horizontal_slider.setOrientation(QtCore.Qt.Horizontal)
            horizontal_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
            horizontal_slider.setEnabled(False)
            horizontal_slider.valueChanged.connect(functools.partial(self._on_slide_trick_slider, trick))
            self._slider_for_trick[trick] = horizontal_slider
            slider_layout.addWidget(horizontal_slider, 0, 1, 1, 10)

            used_difficulties = difficulties_for_trick(self.game_description, trick)
            for i, trick_level in enumerate(iterate_enum(LayoutTrickLevel)):
                if trick_level == LayoutTrickLevel.DISABLED or trick_level in used_difficulties:
                    difficulty_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
                    difficulty_label.setAlignment(QtCore.Qt.AlignHCenter)
                    difficulty_label.setText(trick_level.long_name)

                    slider_layout.addWidget(difficulty_label, 1, 2 * i, 1, 2)

            self.trick_difficulties_layout.addLayout(slider_layout, row, 2, 1, 1)

            if self._window_manager is not None:
                tool_button = QtWidgets.QToolButton(self.trick_level_scroll_contents)
                tool_button.setText("?")
                tool_button.clicked.connect(functools.partial(self._open_trick_details_popup, trick))
                self.trick_difficulties_layout.addWidget(tool_button, row, 3, 1, 1)

            row += 1

        self.trick_level_layout.addLayout(self.trick_difficulties_layout)

    def _on_slide_trick_slider(self, trick: TrickResourceInfo, value: int):
        if self._slider_for_trick[trick].isEnabled():
            with self._editor as options:
                options.set_configuration_field(
                    "trick_level",
                    options.configuration.trick_level.set_level_for_trick(
                        trick,
                        LayoutTrickLevel.from_number(value)
                    )
                )

    def _on_trick_level_minimal_logic_check(self, state: int):
        with self._editor as options:
            options.set_configuration_field(
                "trick_level",
                dataclasses.replace(options.configuration.trick_level,
                                    minimal_logic=bool(state))
            )

    def _exec_trick_details(self, popup: TrickDetailsPopup):
        self._trick_details_popup = popup
        self._trick_details_popup.setWindowModality(Qt.WindowModal)
        self._trick_details_popup.open()

    def _open_trick_details_popup(self, trick: TrickResourceInfo):
        self._exec_trick_details(TrickDetailsPopup(
            self,
            self._window_manager,
            self.game_description,
            trick,
            self._editor.configuration.trick_level.level_for_trick(trick),
        ))

    # Damage strictness
    def setup_damage_elements(self):
        self.damage_strictness_combo.setItemData(0, LayoutDamageStrictness.STRICT)
        self.damage_strictness_combo.setItemData(1, LayoutDamageStrictness.MEDIUM)
        self.damage_strictness_combo.setItemData(2, LayoutDamageStrictness.LENIENT)

        self.damage_strictness_combo.options_field_name = "layout_configuration_damage_strictness"
        self.damage_strictness_combo.currentIndexChanged.connect(functools.partial(_update_options_by_value,
                                                                                   self._editor,
                                                                                   self.damage_strictness_combo))

        def _persist_float(attribute_name: str):
            def persist(value: float):
                with self._editor as options:
                    options.set_configuration_field(attribute_name, value)

            return persist

        self.energy_tank_capacity_spin_box.valueChanged.connect(self._persist_tank_capacity)
        self.dangerous_tank_check.stateChanged.connect(self._persist_dangerous_tank)

        if self.game_enum == RandovaniaGame.PRIME2:
            config_fields = {
                field.name: field
                for field in dataclasses.fields(EchoesConfiguration)
            }
            self.varia_suit_spin_box.setMinimum(config_fields["varia_suit_damage"].metadata["min"])
            self.varia_suit_spin_box.setMaximum(config_fields["varia_suit_damage"].metadata["max"])
            self.dark_suit_spin_box.setMinimum(config_fields["dark_suit_damage"].metadata["min"])
            self.dark_suit_spin_box.setMaximum(config_fields["dark_suit_damage"].metadata["max"])

            self.safe_zone_logic_heal_check.stateChanged.connect(self._persist_safe_zone_logic_heal)
            self.safe_zone_regen_spin.valueChanged.connect(self._persist_safe_zone_regen)
            self.varia_suit_spin_box.valueChanged.connect(_persist_float("varia_suit_damage"))
            self.dark_suit_spin_box.valueChanged.connect(_persist_float("dark_suit_damage"))
        else:
            self.dark_aether_box.setVisible(False)
            self.safe_zone_box.setVisible(False)

    def _persist_tank_capacity(self):
        with self._editor as editor:
            editor.set_configuration_field("energy_per_tank", self.energy_tank_capacity_spin_box.value())

    def _persist_safe_zone_regen(self):
        with self._editor as editor:
            safe_zone = dataclasses.replace(
                editor.configuration.safe_zone,
                heal_per_second=self.safe_zone_regen_spin.value()
            )
            editor.set_configuration_field("safe_zone", safe_zone)

    def _persist_safe_zone_logic_heal(self):
        with self._editor as editor:
            safe_zone = dataclasses.replace(
                editor.configuration.safe_zone,
                fully_heal=self.safe_zone_logic_heal_check.isChecked()
            )
            editor.set_configuration_field("safe_zone", safe_zone)

    def _persist_dangerous_tank(self):
        with self._editor as editor:
            editor.set_configuration_field("dangerous_energy_tank", self.dangerous_tank_check.isChecked())

    # Elevator
    def setup_elevator_elements(self):
        self.elevators_combo.setItemData(0, LayoutElevators.VANILLA)
        self.elevators_combo.setItemData(1, LayoutElevators.TWO_WAY_RANDOMIZED)
        self.elevators_combo.setItemData(2, LayoutElevators.TWO_WAY_UNCHECKED)
        self.elevators_combo.setItemData(3, LayoutElevators.ONE_WAY_ELEVATOR)
        self.elevators_combo.setItemData(4, LayoutElevators.ONE_WAY_ANYTHING)

        self.elevators_combo.options_field_name = "layout_configuration_elevators"
        self.elevators_combo.currentIndexChanged.connect(functools.partial(_update_options_by_value,
                                                                           self._editor,
                                                                           self.elevators_combo))

        if self.game_enum == RandovaniaGame.PRIME3:
            self.patches_tab_widget.setTabText(self.patches_tab_widget.indexOf(self.elevator_tab),
                                               "Teleporters")
            self.elevators_description_label.setText(
                self.elevators_description_label.text().replace("elevator", "teleporter")
            )

    # Starting Area
    def setup_starting_area_elements(self):
        game_description = self.game_description
        world_to_group = {}
        self._starting_location_for_world = {}
        self._starting_location_for_area = {}

        worlds = [
            world
            for world in game_description.world_list.worlds
            if any(area.valid_starting_location and area.default_node_index is not None
                   for area in world.areas)
        ]
        worlds.sort(key=lambda it: it.name)

        for row, world in enumerate(worlds):
            for column, is_dark_world in enumerate(dark_world_flags(world)):
                group_box = QGroupBox(self.starting_locations_contents)
                world_check = QtWidgets.QCheckBox(group_box)
                world_check.setText(world.correct_name(is_dark_world))
                world_check.world_asset_id = world.world_asset_id
                world_check.is_dark_world = is_dark_world
                world_check.stateChanged.connect(functools.partial(self._on_check_starting_world, world_check))
                world_check.setTristate(True)
                vertical_layout = QVBoxLayout(group_box)
                vertical_layout.setContentsMargins(8, 4, 8, 4)
                vertical_layout.setSpacing(2)
                vertical_layout.setAlignment(QtCore.Qt.AlignTop)
                separator = QtWidgets.QFrame()
                separator.setFrameShape(QtWidgets.QFrame.HLine)
                separator.setFrameShadow(QtWidgets.QFrame.Sunken)
                group_box.vertical_layout = vertical_layout
                group_box.vertical_layout.addWidget(world_check)
                group_box.vertical_layout.addWidget(separator)

                world_to_group[world.correct_name(is_dark_world)] = group_box
                self.starting_locations_layout.addWidget(group_box, row, column)
                self._starting_location_for_world[world.correct_name(is_dark_world)] = world_check

        for world in worlds:
            for area in sorted(world.areas, key=lambda a: a.name):
                if not area.valid_starting_location or area.default_node_index is None:
                    continue
                group_box = world_to_group[world.correct_name(area.in_dark_aether)]
                check = QtWidgets.QCheckBox(group_box)
                check.setText(area.name)
                check.area_location = AreaLocation(world.world_asset_id, area.area_asset_id)
                check.stateChanged.connect(functools.partial(self._on_check_starting_area, check))
                group_box.vertical_layout.addWidget(check)
                self._starting_location_for_area[area.area_asset_id] = check

        self.starting_area_quick_fill_ship.clicked.connect(self._starting_location_on_select_ship)
        self.starting_area_quick_fill_save_station.clicked.connect(self._starting_location_on_select_save_station)

    def _on_check_starting_area(self, check, _):
        if self._during_batch_check_update:
            return
        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                editor.configuration.starting_location.ensure_has_location(check.area_location,
                                                                           check.isChecked())
            )

    def _on_check_starting_world(self, check, _):
        if self._during_batch_check_update:
            return
        world_list = self.game_description.world_list
        world = world_list.world_by_asset_id(check.world_asset_id)
        world_areas = [world_list.area_to_area_location(area)
                       for area in world.areas if check.is_dark_world == area.in_dark_aether
                       if area.area_asset_id in self._starting_location_for_area]
        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                editor.configuration.starting_location.ensure_has_locations(world_areas, check.isChecked())
            )

    def _starting_location_on_select_ship(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                StartingLocation.with_elements([self.game_description.starting_location], self.game_enum)
            )

    def _starting_location_on_select_save_station(self):
        world_list = self.game_description.world_list
        save_stations = [world_list.node_to_area_location(node)
                         for node in world_list.all_nodes if node.name == "Save Station"]

        with self._editor as editor:
            editor.set_configuration_field(
                "starting_location",
                StartingLocation.with_elements(save_stations, self.game_enum)
            )

    def on_preset_changed_starting_area(self, preset: Preset):
        starting_locations = preset.configuration.starting_location.locations

        self._during_batch_check_update = True
        for world in self.game_description.world_list.worlds:
            for is_dark_world in dark_world_flags(world):
                all_areas = True
                no_areas = True
                areas = [area for area in world.areas if area.in_dark_aether == is_dark_world]
                correct_name = world.correct_name(is_dark_world)
                if correct_name not in self._starting_location_for_world:
                    continue

                for area in areas:
                    if area.area_asset_id in self._starting_location_for_area:
                        is_checked = AreaLocation(world.world_asset_id, area.area_asset_id) in starting_locations
                        if is_checked:
                            no_areas = False
                        else:
                            all_areas = False
                        self._starting_location_for_area[area.area_asset_id].setChecked(is_checked)
                if all_areas:
                    self._starting_location_for_world[correct_name].setCheckState(Qt.Checked)
                elif no_areas:
                    self._starting_location_for_world[correct_name].setCheckState(Qt.Unchecked)
                else:
                    self._starting_location_for_world[correct_name].setCheckState(Qt.PartiallyChecked)
        self._during_batch_check_update = False

    # Location Pool
    def setup_location_pool_elements(self):
        self.randomization_mode_combo.setItemData(0, RandomizationMode.FULL)
        self.randomization_mode_combo.setItemData(1, RandomizationMode.MAJOR_MINOR_SPLIT)
        self.randomization_mode_combo.currentIndexChanged.connect(self._on_update_randomization_mode)

        vertical_layouts = [
            QtWidgets.QVBoxLayout(self.excluded_locations_area_contents),
            QtWidgets.QVBoxLayout(self.excluded_locations_area_contents),
        ]
        for layout in vertical_layouts:
            self.excluded_locations_area_layout.addLayout(layout)

        world_list = self.game_description.world_list
        self._location_pool_for_node = {}

        nodes_by_world = collections.defaultdict(list)
        node_names = {}
        pickup_match = re.compile(r"Pickup \(([^\)]+)\)")

        for world in world_list.worlds:
            for is_dark_world in dark_world_flags(world):
                for area in world.areas:
                    if area.in_dark_aether != is_dark_world:
                        continue
                    for node in area.nodes:
                        if isinstance(node, PickupNode):
                            nodes_by_world[world.correct_name(is_dark_world)].append(node)
                            match = pickup_match.match(node.name)
                            if match is not None:
                                node_name = match.group(1)
                            else:
                                node_name = node.name
                            node_names[node] = f"{world_list.nodes_to_area(node).name} ({node_name})"

        for i, world_name in enumerate(sorted(nodes_by_world.keys())):
            group_box = QGroupBox(self.excluded_locations_area_contents)
            group_box.setTitle(world_name)
            vertical_layout = QVBoxLayout(group_box)
            vertical_layout.setContentsMargins(8, 4, 8, 4)
            vertical_layout.setSpacing(2)
            group_box.vertical_layout = vertical_layout
            vertical_layouts[i % len(vertical_layouts)].addWidget(group_box)

            for node in sorted(nodes_by_world[world_name], key=node_names.get):
                check = QtWidgets.QCheckBox(group_box)
                check.setText(node_names[node])
                check.node = node
                check.stateChanged.connect(functools.partial(self._on_check_location, check))
                group_box.vertical_layout.addWidget(check)
                self._location_pool_for_node[node] = check

        for layout in vertical_layouts:
            layout.addSpacerItem(
                QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding))

    def _on_update_randomization_mode(self):
        with self._editor as editor:
            editor.available_locations = dataclasses.replace(
                editor.available_locations, randomization_mode=self.randomization_mode_combo.currentData())

    def _on_check_location(self, check: QtWidgets.QCheckBox, _):
        if self._during_batch_check_update:
            return
        with self._editor as editor:
            editor.available_locations = editor.available_locations.ensure_index(check.node.pickup_index,
                                                                                 not check.isChecked())



from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class PickupModelStyle(BitPackEnum, Enum):
    ALL_VISIBLE = "all-visible"
    HIDE_MODEL = "hide-model"
    HIDE_SCAN = "hide-scan"
    HIDE_ALL = "hide-all"

    @classmethod
    def default(cls) -> "PickupModelStyle":
        return cls.ALL_VISIBLE

    @property
    def long_name(self):
        return MODEL_TYPE_PRETTY[self]


MODEL_TYPE_PRETTY = {
    PickupModelStyle.ALL_VISIBLE: "All visible",
    PickupModelStyle.HIDE_MODEL: "Hide model",
    PickupModelStyle.HIDE_SCAN: "Hide model and scan",
    PickupModelStyle.HIDE_ALL: "Hide all",
}


class PickupModelDataSource(BitPackEnum, Enum):
    ETM = "etm"
    RANDOM = "random"
    LOCATION = "location"

    @classmethod
    def default(cls) -> "PickupModelDataSource":
        return cls.ETM

    @property
    def long_name(self):
        return DATA_SOURCE_PRETTY[self]


DATA_SOURCE_PRETTY = {
    PickupModelDataSource.ETM: "ETM",
    PickupModelDataSource.RANDOM: "Random",
    PickupModelDataSource.LOCATION: "Vanilla",
}



from randovania.layout.permalink import Permalink


class GenerationFailure(Exception):
    permalink: Permalink
    source: Exception

    def __init__(self, reason: str, permalink: Permalink, source: Exception):
        super().__init__(reason)
        self.permalink = permalink
        self.source = source

    def __reduce__(self):
        return GenerationFailure, (super().__str__(), self.permalink, self.source)

    def __eq__(self, other):
        if not isinstance(other, GenerationFailure):
            return False

        if self.permalink != other.permalink:
            return False

        return super(Exception, other).__str__() == super().__str__()


class ImpossibleForSolver(Exception):
    pass


class InvalidConfiguration(Exception):
    pass



import dataclasses
from enum import Enum
from typing import Tuple, Iterator, FrozenSet, List

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackValue, BitPackDecoder, BitPackEnum
from randovania.game_description.resources.pickup_index import PickupIndex

_ALL_INDICES = list(range(119))


class RandomizationMode(BitPackEnum, Enum):
    FULL = "full"
    MAJOR_MINOR_SPLIT = "major/minor split"

    @classmethod
    def default(cls) -> "RandomizationMode":
        return cls.FULL


@dataclasses.dataclass(frozen=True)
class AvailableLocationsConfiguration(BitPackValue):
    randomization_mode: RandomizationMode
    excluded_indices: FrozenSet[PickupIndex]

    @property
    def _sorted_indices(self) -> List[int]:
        return list(sorted(item.index for item in self.excluded_indices))

    @property
    def as_json(self) -> dict:
        return {
            "randomization_mode": self.randomization_mode.value,
            "excluded_indices": self._sorted_indices,
        }

    @classmethod
    def from_json(cls, value: dict) -> "AvailableLocationsConfiguration":
        return cls(
            randomization_mode=RandomizationMode(value["randomization_mode"]),
            excluded_indices=frozenset(PickupIndex(item) for item in value["excluded_indices"]),
        )

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        yield from self.randomization_mode.bit_pack_encode(metadata)
        if self.excluded_indices:
            yield from bitpacking.encode_bool(True)
            yield from bitpacking.pack_sorted_array_elements(self._sorted_indices, _ALL_INDICES)
        else:
            yield from bitpacking.encode_bool(False)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        randomization_mode = RandomizationMode.bit_pack_unpack(decoder, metadata)

        if bitpacking.decode_bool(decoder):
            indices = bitpacking.decode_sorted_array_elements(decoder, _ALL_INDICES)
        else:
            indices = []

        return AvailableLocationsConfiguration(
            randomization_mode=randomization_mode,
            excluded_indices=frozenset(PickupIndex(item) for item in indices),
        )

    def ensure_index(self, index: PickupIndex, present: bool):
        excluded_indices = set(self.excluded_indices)
        if present:
            excluded_indices.add(index)
        elif index in excluded_indices:
            excluded_indices.remove(index)
        return dataclasses.replace(self, excluded_indices=frozenset(excluded_indices))



import shutil
from pathlib import Path

from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.prime import iso_packager, claris_randomizer
from randovania.games.prime.banner_patcher import patch_game_name_and_id
from randovania.interface_common import echoes
from randovania.interface_common.options import Options
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.interface_common.status_update_lib import ProgressUpdateCallable, ConstantPercentageCallback
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink

export_busy = False


def delete_files_location(options: Options, ):
    """
    Deletes an extracted game in given options.
    :param options:
    :return:
    """
    game_files_path = options.game_files_path
    if game_files_path.exists():
        shutil.rmtree(game_files_path)

    backup_files_path = options.backup_files_path
    if backup_files_path.exists():
        shutil.rmtree(backup_files_path)


def unpack_iso(input_iso: Path,
               options: Options,
               progress_update: ProgressUpdateCallable,
               ):
    """
    Unpacks the given ISO to the files listed in options
    :param input_iso:
    :param options:
    :param progress_update:
    :return:
    """
    game_files_path = options.game_files_path

    delete_files_location(options)
    iso_packager.unpack_iso(
        iso=input_iso,
        game_files_path=game_files_path,
        progress_update=progress_update,
    )


def generate_layout(options: Options,
                    permalink: Permalink,
                    progress_update: ProgressUpdateCallable,
                    ) -> LayoutDescription:
    """
    Creates a LayoutDescription for the configured permalink
    :param options:
    :param permalink:
    :param progress_update:
    :return:
    """
    return echoes.generate_layout(
        permalink=permalink,
        status_update=ConstantPercentageCallback(progress_update, -1),
        validate_after_generation=options.advanced_validate_seed_after,
        timeout_during_generation=options.advanced_timeout_during_generation,
    )


def apply_layout(layout: LayoutDescription,
                 options: Options,
                 players_config: PlayersConfiguration,
                 progress_update: ProgressUpdateCallable):
    """
    Applies the given LayoutDescription to the files listed in options
    :param options:
    :param layout:
    :param players_config:
    :param progress_update:
    :return:
    """
    game_files_path = options.game_files_path
    backup_files_path = options.backup_files_path

    patch_game_name_and_id(game_files_path, "Metroid Prime 2: Randomizer - {}".format(layout.shareable_hash))

    claris_randomizer.apply_layout(description=layout,
                                   players_config=players_config,
                                   cosmetic_patches=options.cosmetic_patches,
                                   backup_files_path=backup_files_path,
                                   progress_update=progress_update,
                                   game_root=game_files_path,
                                   )


def apply_patcher_file(patcher_file: dict,
                       game_specific: EchoesGameSpecific,
                       shareable_hash: str,
                       options: Options,
                       progress_update: ProgressUpdateCallable):
    """
    Applies the given LayoutDescription to the files listed in options
    :param options:
    :param patcher_file:
    :param game_specific:
    :param shareable_hash:
    :param progress_update:
    :return:
    """
    game_files_path = options.game_files_path
    backup_files_path = options.backup_files_path

    patch_game_name_and_id(game_files_path, "Metroid Prime 2: Randomizer - {}".format(shareable_hash))

    claris_randomizer.apply_patcher_file(
        game_root=game_files_path,
        backup_files_path=backup_files_path,
        patcher_data=patcher_file,
        game_specific=game_specific,
        progress_update=progress_update,
    )


def pack_iso(output_iso: Path,
             options: Options,
             progress_update: ProgressUpdateCallable,
             ):
    """
    Unpacks the files listed in options to the given path
    :param output_iso:
    :param options:
    :param progress_update:
    :return:
    """
    game_files_path = options.game_files_path

    iso_packager.pack_iso(
        iso=output_iso,
        game_files_path=game_files_path,
        progress_update=progress_update,
    )



from unittest.mock import MagicMock

import pytest

from randovania.gui.dialog.permalink_dialog import PermalinkDialog


@pytest.mark.parametrize("valid", [False, True])
def test_on_permalink_changed(skip_qtbot, mocker, valid):
    mock_from_str: MagicMock = mocker.patch("randovania.layout.permalink.Permalink.from_str")
    mock_from_str.return_value.as_base64_str = ""
    if not valid:
        mock_from_str.side_effect = ValueError("Invalid permalink")

    dialog = PermalinkDialog()
    dialog.permalink_edit.setText("")
    skip_qtbot.addWidget(dialog)

    # Run
    dialog._on_permalink_changed("")

    # Assert
    assert dialog.accept_button.isEnabled() == valid


def test_on_permalink_changed_permalink_different_str(skip_qtbot, mocker):
    mock_from_str: MagicMock = mocker.patch("randovania.layout.permalink.Permalink.from_str")
    mock_from_str.return_value.as_base64_str = "!permalink_wrong"

    dialog = PermalinkDialog()
    skip_qtbot.addWidget(dialog)

    # Run
    dialog._on_permalink_changed("")

    # Assert
    assert not dialog.accept_button.isEnabled()
    assert dialog.import_error_label.text() == "Invalid permalink: Imported permalink is different from text field."



from mock import MagicMock

import pytest

from randovania.game_connection.connection_base import GameConnectionStatus, InventoryItem
from randovania.gui.auto_tracker_window import AutoTrackerWindow


@pytest.fixture(name="window")
def auto_tracker_window(skip_qtbot):
    connection = MagicMock()
    connection.pretty_current_status = "Pretty"
    return AutoTrackerWindow(connection, MagicMock())


def test_update_tracker_from_hook(window, echoes_resource_database):
    # Setup
    inventory = {
        item: InventoryItem(item.index % 3, item.index % 3)
        for item in echoes_resource_database.item
    }

    # Run
    window._update_tracker_from_hook(inventory)


@pytest.mark.parametrize("current_status", [GameConnectionStatus.Disconnected,
                                            GameConnectionStatus.TrackerOnly,
                                            GameConnectionStatus.InGame])
@pytest.mark.asyncio
async def test_on_timer_update(current_status: GameConnectionStatus,
                               window):
    # Setup
    inventory = {}
    game_connection = MagicMock()
    game_connection.pretty_current_status = "Pretty Status"

    window = AutoTrackerWindow(game_connection, MagicMock())
    window._update_tracker_from_hook = MagicMock()

    game_connection.get_current_inventory.return_value = inventory
    game_connection.current_status = current_status

    # Run
    await window._on_timer_update()

    # Assert
    if current_status != GameConnectionStatus.Disconnected:
        window._update_tracker_from_hook.assert_called_once_with(inventory)
    else:
        window._update_tracker_from_hook.assert_not_called()



from typing import Optional, List, Tuple

from PySide2.QtCore import Qt
from PySide2.QtGui import QIntValidator
from PySide2.QtWidgets import QPushButton, QWidget, QGroupBox, QVBoxLayout, QLabel, QGridLayout, QHBoxLayout, QComboBox, \
    QLineEdit

import randovania.game_description.pretty_print
from randovania.game_description import data_writer
from randovania.game_description.requirements import ResourceRequirement, Requirement
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.resource_type import ResourceType


def _create_resource_name_combo(resource_database: ResourceDatabase,
                                resource_type: ResourceType,
                                current_resource: Optional[ResourceInfo],
                                parent: QWidget,
                                ) -> QComboBox:
    """

    :param resource_database:
    :param current_resource:
    :param parent:
    :return:
    """

    resource_name_combo = QComboBox(parent)

    for resource in sorted(resource_database.get_by_type(resource_type), key=lambda x: x.long_name):
        resource_name_combo.addItem(resource.long_name, resource)
        if resource is current_resource:
            resource_name_combo.setCurrentIndex(resource_name_combo.count() - 1)

    return resource_name_combo


def _create_resource_type_combo(current_resource_type: ResourceType, parent: QWidget) -> QComboBox:
    """

    :param current_resource_type:
    :param parent:
    :return:
    """
    resource_type_combo = QComboBox(parent)

    for resource_type in ResourceType:
        if not resource_type.is_usable_for_requirement:
            continue

        resource_type_combo.addItem(resource_type.name, resource_type)
        if resource_type is current_resource_type:
            resource_type_combo.setCurrentIndex(resource_type_combo.count() - 1)

    return resource_type_combo


class ItemRow:
    def __init__(self,
                 parent: QWidget, parent_layout: QVBoxLayout,
                 resource_database: ResourceDatabase, item: ResourceRequirement,
                 rows: List["ItemRow"]
                 ):
        self.parent = parent
        self.resource_database = resource_database
        self._rows = rows
        rows.append(self)

        self.layout = QHBoxLayout()
        self.layout.setObjectName(f"Box layout for {item.resource.long_name}")
        parent_layout.addLayout(self.layout)

        self.resource_type_combo = _create_resource_type_combo(item.resource.resource_type, parent)
        self.resource_type_combo.setMinimumWidth(75)
        self.resource_type_combo.setMaximumWidth(75)

        self.resource_name_combo = _create_resource_name_combo(self.resource_database,
                                                               item.resource.resource_type,
                                                               item.resource,
                                                               self.parent)

        self.negate_combo = QComboBox(parent)
        self.negate_combo.addItem("≥", False)
        self.negate_combo.addItem("<", True)
        self.negate_combo.setCurrentIndex(int(item.negate))
        self.negate_combo.setMinimumWidth(40)
        self.negate_combo.setMaximumWidth(40)

        self.amount_edit = QLineEdit(parent)
        self.amount_edit.setValidator(QIntValidator(1, 10000))
        self.amount_edit.setText(str(item.amount))
        self.amount_edit.setMinimumWidth(45)
        self.amount_edit.setMaximumWidth(45)

        self.remove_button = QPushButton(parent)
        self.remove_button.setText("X")
        self.remove_button.setMaximumWidth(20)

        self.layout.addWidget(self.resource_type_combo)
        self.layout.addWidget(self.resource_name_combo)
        self.layout.addWidget(self.negate_combo)
        self.layout.addWidget(self.amount_edit)
        self.layout.addWidget(self.remove_button)

        self.resource_type_combo.currentIndexChanged.connect(self._update_type)
        self.remove_button.clicked.connect(self._delete_row)

    def _update_type(self):
        old_combo = self.resource_name_combo
        self.resource_name_combo = _create_resource_name_combo(self.resource_database,
                                                               self.resource_type_combo.currentData(),
                                                               None,
                                                               self.parent)

        self.layout.replaceWidget(old_combo, self.resource_name_combo)
        old_combo.deleteLater()

    def _delete_row(self):
        self.resource_type_combo.deleteLater()
        self.resource_name_combo.deleteLater()
        self.negate_combo.deleteLater()
        self.amount_edit.deleteLater()
        self.remove_button.deleteLater()
        self.layout.deleteLater()
        self._rows.remove(self)

    @property
    def current_individual(self) -> ResourceRequirement:
        return ResourceRequirement(
            self.resource_name_combo.currentData(),
            int(self.amount_edit.text()),
            self.negate_combo.currentData()
        )


class ConnectionsVisualizer:
    parent: QWidget
    resource_database: ResourceDatabase
    grid_layout: QGridLayout
    _elements: List[QWidget]

    def __init__(self,
                 parent: QWidget,
                 grid_layout: QGridLayout,
                 resource_database: ResourceDatabase,
                 requirement: Requirement,
                 edit_mode: bool
                 ):
        self.parent = parent
        self.resource_database = resource_database
        self.edit_mode = edit_mode
        self.grid_layout = grid_layout
        self._elements = []

        self._add_widget_for_requirement_array(requirement)

    def _add_widget_for_requirement_array(self, requirement: Requirement):
        self.grid_layout.setAlignment(Qt.AlignTop)

        parents: List[Tuple[QGroupBox, QVBoxLayout]] = [(self.parent, self.grid_layout)]

        for depth, text in randovania.game_description.pretty_print.pretty_print_requirement(requirement):
            if "of the following" in text:
                parent = parents[depth]
                group_box = QGroupBox(parent[0])
                group_box.setContentsMargins(8, 0, 2, 6)
                group_box.setTitle(text)
                self._elements.append(group_box)
                vertical_layout = QVBoxLayout(group_box)
                vertical_layout.setAlignment(Qt.AlignTop)
                parent[1].addWidget(group_box)
                if len(parents) <= depth + 1:
                    parents.append(None)
                parents[depth + 1] = (group_box, vertical_layout)
            else:
                label = QLabel(parents[depth][0])
                label.setText(text)
                self._elements.append(label)
                parents[depth][1].addWidget(label)

    def deleteLater(self):
        for element in self._elements:
            element.deleteLater()



import dataclasses
import json
import re
from pathlib import Path
from typing import Dict, Optional

from PySide2 import QtGui
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QMainWindow, QRadioButton, QGridLayout, QDialog, QFileDialog, QInputDialog, QMessageBox
from asyncqt import asyncSlot

from randovania.game_description import data_reader, data_writer, pretty_print
from randovania.game_description.area import Area
from randovania.game_description.node import Node, DockNode, TeleporterNode, GenericNode
from randovania.game_description.requirements import Requirement
from randovania.game_description.world import World
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.gui.dialog.connections_editor import ConnectionsEditor
from randovania.gui.dialog.node_details_popup import NodeDetailsPopup
from randovania.gui.generated.data_editor_ui import Ui_DataEditorWindow
from randovania.gui.lib import async_dialog
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.gui.lib.connections_visualizer import ConnectionsVisualizer


class DataEditorWindow(QMainWindow, Ui_DataEditorWindow):
    edit_mode: bool
    selected_node_button: QRadioButton = None
    radio_button_to_node: Dict[QRadioButton, Node]
    _area_with_displayed_connections: Optional[Area] = None
    _previous_selected_node: Optional[Node] = None
    _connections_visualizer: Optional[ConnectionsVisualizer] = None
    _edit_popup: Optional[QDialog] = None

    def __init__(self, data: dict, data_path: Optional[Path], is_internal: bool, edit_mode: bool):
        super().__init__()
        self.setupUi(self)
        set_default_window_icon(self)

        self._is_internal = is_internal
        self._data_path = data_path
        self._last_data = data
        self.edit_mode = edit_mode
        self.radio_button_to_node = {}

        self.world_selector_box.currentIndexChanged.connect(self.on_select_world)
        self.area_selector_box.currentIndexChanged.connect(self.on_select_area)
        self.node_details_label.linkActivated.connect(self._on_click_link_to_other_node)
        self.node_heals_check.stateChanged.connect(self.on_node_heals_check)
        self.area_spawn_check.stateChanged.connect(self.on_area_spawn_check)
        self.node_edit_button.clicked.connect(self.on_node_edit_button)
        self.other_node_connection_edit_button.clicked.connect(self._open_edit_connection)

        self.save_database_button.setEnabled(data_path is not None)
        if self._is_internal:
            self.save_database_button.clicked.connect(self._save_as_internal_database)
        else:
            self.save_database_button.clicked.connect(self._prompt_save_database)

        self.new_node_button.clicked.connect(self._create_new_node)
        self.delete_node_button.clicked.connect(self._remove_node)
        self.verticalLayout.setAlignment(Qt.AlignTop)
        self.alternatives_grid_layout = QGridLayout(self.other_node_alternatives_contents)

        world_reader, self.game_description = data_reader.decode_data_with_world_reader(data)
        self.generic_index = world_reader.generic_index

        self.resource_database = self.game_description.resource_database
        self.world_list = self.game_description.world_list

        for world in sorted(self.world_list.worlds, key=lambda x: x.name):
            name = "{0.name} ({0.dark_name})".format(world) if world.dark_name else world.name
            self.world_selector_box.addItem(name, userData=world)

        self.update_edit_mode()

    @classmethod
    def open_internal_data(cls, game: RandovaniaGame, edit_mode: bool) -> "DataEditorWindow":
        default_data.read_json_then_binary.cache_clear()
        path, data = default_data.read_json_then_binary(game)
        if path.suffix == ".bin":
            path = None
        return DataEditorWindow(data, path, True, edit_mode)

    def closeEvent(self, event: QtGui.QCloseEvent):
        if self._check_for_edit_dialog():
            event.ignore()
        else:
            data = data_writer.write_game_description(self.game_description)
            if data != self._last_data:
                user_response = QMessageBox.warning(self, "Unsaved changes",
                                                    "You have unsaved changes. Do you want to close and discard?",
                                                    QMessageBox.Yes | QMessageBox.No,
                                                    QMessageBox.No)
                if user_response == QMessageBox.No:
                    return event.ignore()
            super().closeEvent(event)

    def on_select_world(self):
        self.area_selector_box.clear()
        for area in sorted(self.current_world.areas, key=lambda x: x.name):
            if area.name.startswith("!!"):
                continue
            self.area_selector_box.addItem(area.name, userData=area)
        self.area_selector_box.setEnabled(True)

    def on_select_area(self):
        for node in self.radio_button_to_node.keys():
            node.deleteLater()

        self.radio_button_to_node.clear()

        current_area = self.current_area
        if not current_area:
            self.new_node_button.setEnabled(False)
            self.delete_node_button.setEnabled(False)
            return

        is_first = True
        for node in sorted(current_area.nodes, key=lambda x: x.name):
            button = QRadioButton(self.points_of_interest_group)
            button.setText(node.name)
            self.radio_button_to_node[button] = node
            if is_first:
                self.selected_node_button = button

            button.setChecked(is_first)
            button.toggled.connect(self.on_select_node)
            is_first = False
            self.verticalLayout.addWidget(button)

        self.new_node_button.setEnabled(True)
        self.delete_node_button.setEnabled(len(current_area.nodes) > 1)

        self.update_selected_node()

    def on_select_node(self, active):
        if active:
            self.selected_node_button = self.sender()
            self.update_selected_node()

    def focus_on_world(self, world_name: str):
        world = self.world_list.world_with_name(world_name)
        self.world_selector_box.setCurrentIndex(self.world_selector_box.findData(world))

    def focus_on_area(self, area_name: str):
        self.area_selector_box.setCurrentIndex(self.area_selector_box.findText(area_name))

    def _on_click_link_to_other_node(self, link: str):
        world_name, area_name, node_name = None, None, None

        info = re.match(r"^node://([^/]+)/([^/]+)/(.+)$", link)
        if info:
            world_name, area_name, node_name = info.group(1, 2, 3)
        else:
            info = re.match(r"^area://([^/]+)/([^/]+)$", link)
            world_name, area_name = info.group(1, 2)

        if world_name is not None and area_name is not None:
            self.focus_on_world(world_name)
            self.focus_on_area(area_name)
            if node_name is None and self.current_area.default_node_index is not None:
                node_name = self.current_area.nodes[self.current_area.default_node_index].name
            
            for radio_button in self.radio_button_to_node.keys():
                if radio_button.text() == node_name:
                    radio_button.setChecked(True)

    def on_node_heals_check(self, state: int):
        old_node = self.current_node
        assert old_node is not None

        new_node = dataclasses.replace(old_node, heal=bool(state))
        self.replace_node_with(self.current_area, old_node, new_node)

    def on_area_spawn_check(self, state: int):
        state = bool(state)
        if not state:
            return

        object.__setattr__(self.current_area, "default_node_index", self.current_area.nodes.index(self.current_node))
        self.area_spawn_check.setEnabled(False)

    def replace_node_with(self, area: Area, old_node: Node, new_node: Node):
        if old_node == new_node:
            return

        def sub(n: Node):
            return new_node if n == old_node else n

        area_node_list = area.nodes
        for i, node in enumerate(area_node_list):
            if node == old_node:
                area_node_list[i] = new_node

        new_connections = {
            sub(source_node): {
                sub(target_node): requirements
                for target_node, requirements in connection.items()
            }
            for source_node, connection in area.connections.items()
        }
        area.connections.clear()
        area.connections.update(new_connections)
        self.game_description.world_list.refresh_node_cache()

        if area == self.current_area:
            radio = next(key for key, value in self.radio_button_to_node.items() if value == old_node)
            radio.setText(new_node.name)
            self.radio_button_to_node[radio] = new_node
            self.update_selected_node()

    def _check_for_edit_dialog(self) -> bool:
        """
        If an edit popup exists, raises it and returns True.
        Otherwise, just return False.
        :return:
        """
        if self._edit_popup is not None:
            self._edit_popup.raise_()
            return True
        else:
            return False

    async def _execute_edit_dialog(self, dialog: QDialog):
        self._edit_popup = dialog
        try:
            result = await async_dialog.execute_dialog(self._edit_popup)
            return result == QDialog.Accepted
        finally:
            self._edit_popup = None

    @asyncSlot()
    async def on_node_edit_button(self):
        if self._check_for_edit_dialog():
            return

        area = self.current_area
        node_edit_popup = NodeDetailsPopup(self.game_description, self.current_node)
        if await self._execute_edit_dialog(node_edit_popup):
            try:
                new_node = node_edit_popup.create_new_node()
            except ValueError as e:
                await async_dialog.warning(self, "Error in new node", str(e))
                return
            self.replace_node_with(area, node_edit_popup.node, new_node)

    def update_selected_node(self):
        node = self.current_node
        assert node is not None

        self.node_heals_check.setChecked(node.heal)
        is_default_spawn = self.current_area.default_node_index == self.current_area.nodes.index(node)
        self.area_spawn_check.setChecked(is_default_spawn)
        self.area_spawn_check.setEnabled(self.edit_mode and not is_default_spawn)

        msg = pretty_print.pretty_print_node_type(node, self.world_list)
        if isinstance(node, DockNode):
            try:
                other = self.world_list.resolve_dock_connection(self.current_world, node.default_connection)
                msg = "{} to <a href=\"node://{}\">{}</a>".format(
                    node.default_dock_weakness.name,
                    self.world_list.node_name(other, with_world=True),
                    self.world_list.node_name(other)
                )
            except IndexError:
                pass

        elif isinstance(node, TeleporterNode):
            other = self.world_list.area_by_area_location(node.default_connection)
            name = self.world_list.area_name(other, separator="/", distinguish_dark_aether=False)
            pretty_name = msg.replace("Teleporter to ", "")
            msg = f'Teleporter to <a href="area://{name}">{pretty_name}</a>'

        self.node_name_label.setText(node.name)
        self.node_details_label.setText(msg)
        self.update_other_node_connection()

        self._previous_selected_node = node

    def update_other_node_connection(self):
        """
        Fills self.other_node_connection_combo for the current area, excluding the currently selected node.
        :return:
        """
        current_node = self.current_node
        assert current_node is not None

        # Calculates which node should be selected
        selected_node = None
        if self._area_with_displayed_connections == self.current_area:
            selected_node = self.current_connection_node
        if selected_node is current_node:
            selected_node = self._previous_selected_node

        #
        self._area_with_displayed_connections = self.current_area

        if self.other_node_connection_combo.count() > 0:
            self.other_node_connection_combo.currentIndexChanged.disconnect(self.update_connections)
            self.other_node_connection_combo.clear()

        for node in sorted(self.current_area.nodes, key=lambda x: x.name):
            if node is not current_node:
                self.other_node_connection_combo.addItem(node.name, userData=node)
                if node is selected_node:
                    self.other_node_connection_combo.setCurrentIndex(self.other_node_connection_combo.count() - 1)

        if self.other_node_connection_combo.count() > 0:
            self.other_node_connection_combo.currentIndexChanged.connect(self.update_connections)
            self.other_node_connection_combo.setEnabled(True)
            self.other_node_connection_edit_button.setEnabled(True)
        else:
            self.other_node_connection_combo.setEnabled(False)
            self.other_node_connection_edit_button.setEnabled(False)

        self.update_connections()

    def update_connections(self):
        current_node = self.current_node
        current_connection_node = self.current_connection_node

        assert current_node is not None
        assert current_node != current_connection_node

        if self._connections_visualizer is not None:
            self._connections_visualizer.deleteLater()
            self._connections_visualizer = None

        if current_connection_node is None:
            assert len(self.current_area.nodes) == 1
            return

        requirement = self.current_area.connections[self.current_node].get(self.current_connection_node,
                                                                           Requirement.impossible())
        self._connections_visualizer = ConnectionsVisualizer(
            self.other_node_alternatives_contents,
            self.alternatives_grid_layout,
            self.resource_database,
            requirement,
            False
        )

    @asyncSlot()
    async def _open_edit_connection(self):
        if self._check_for_edit_dialog():
            return

        from_node = self.current_node
        target_node = self.current_connection_node
        assert from_node is not None
        assert target_node is not None

        requirement = self.current_area.connections[from_node].get(target_node, Requirement.impossible())
        editor = ConnectionsEditor(self, self.resource_database, requirement)
        if await self._execute_edit_dialog(editor):
            self._apply_edit_connections(from_node, target_node, editor.final_requirement)

    def _apply_edit_connections(self, from_node: Node, target_node: Node,
                                requirement: Optional[Requirement]):

        current_connections = self.current_area.connections[from_node]
        self.current_area.connections[from_node][target_node] = requirement
        if self.current_area.connections[from_node][target_node] is None:
            del self.current_area.connections[from_node][target_node]

        self.current_area.connections[from_node] = {
            node: current_connections[node]
            for node in self.current_area.nodes
            if node in current_connections
        }
        self.update_connections()

    def _prompt_save_database(self):
        open_result = QFileDialog.getSaveFileName(self, caption="Select a Randovania database path.", filter="*.json")
        if not open_result or open_result == ("", ""):
            return
        self._save_database(Path(open_result[0]))

    def _save_database(self, path: Path):
        data = data_writer.write_game_description(self.game_description)
        with path.open("w") as open_file:
            json.dump(data, open_file, indent=4)
        self._last_data = data

    def _save_as_internal_database(self):
        self._save_database(self._data_path)
        with self._data_path.with_suffix(".txt").open("w", encoding="utf-8") as output:
            pretty_print.write_human_readable_world_list(self.game_description, output)

    def _create_new_node(self):
        node_name, did_confirm = QInputDialog.getText(self, "New Node", "Insert node name:")
        if not did_confirm or node_name == "":
            return

        if self.current_area.node_with_name(node_name) is not None:
            QMessageBox.warning(self,
                                "New Node",
                                "A node named '{}' already exists.".format(node_name))
            return

        self._do_create_node(node_name)

    def _do_create_node(self, node_name: str):
        self.generic_index += 1
        new_node = GenericNode(node_name, False, None, self.generic_index)
        self.current_area.nodes.append(new_node)
        self.current_area.connections[new_node] = {}
        self.game_description.world_list.refresh_node_cache()

        self.on_select_area()

    def _remove_node(self):
        if self._check_for_edit_dialog():
            return

        current_node = self.current_node

        if not isinstance(current_node, GenericNode):
            QMessageBox.warning(self, "Delete Node", "Can only remove Generic Nodes")
            return

        user_response = QMessageBox.warning(
            self,
            "Delete Node",
            "Are you sure you want to delete the node '{}'?".format(current_node.name),
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No
        )

        if user_response != QMessageBox.Yes:
            return

        for other_nodes in self.current_area.connections.values():
            other_nodes.pop(current_node, None)

        self.current_area.remove_node(current_node)
        self.game_description.world_list.refresh_node_cache()
        self.on_select_area()

    def update_edit_mode(self):
        self.delete_node_button.setVisible(self.edit_mode)
        self.new_node_button.setVisible(self.edit_mode)
        self.save_database_button.setVisible(self.edit_mode)
        self.other_node_connection_edit_button.setVisible(self.edit_mode)
        self.node_heals_check.setEnabled(self.edit_mode)
        self.area_spawn_check.setEnabled(self.edit_mode and self.area_spawn_check.isEnabled())
        self.node_edit_button.setVisible(self.edit_mode)

    @property
    def current_world(self) -> World:
        return self.world_selector_box.currentData()

    @property
    def current_area(self) -> Area:
        return self.area_selector_box.currentData()

    @property
    def current_node(self) -> Optional[Node]:
        return self.radio_button_to_node.get(self.selected_node_button)

    @property
    def current_connection_node(self) -> Node:
        return self.other_node_connection_combo.currentData()



from typing import List, Tuple, Optional, Dict

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources, add_resource_gain_to_current_resources
from randovania.generator.item_pool import PoolResults
from randovania.generator.item_pool.pickup_creator import create_major_item
from randovania.layout.ammo_configuration import AmmoConfiguration
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.resolver.exceptions import InvalidConfiguration


def _find_ammo_for(ammo_index: Tuple[int, ...],
                   ammo_configuration: AmmoConfiguration,
                   ) -> Tuple[Optional[Ammo], bool]:
    for ammo, ammo_state in ammo_configuration.items_state.items():
        if ammo.items == ammo_index:
            return ammo, ammo_state.requires_major_item

    return None, False


def add_major_items(resource_database: ResourceDatabase,
                    major_items_configuration: MajorItemsConfiguration,
                    ammo_configuration: AmmoConfiguration,
                    ) -> PoolResults:
    """

    :param resource_database:
    :param major_items_configuration:
    :param ammo_configuration:
    :return:
    """

    item_pool: List[PickupEntry] = []
    new_assignment: Dict[PickupIndex, PickupEntry] = {}
    initial_resources: CurrentResources = {}

    for item, state in major_items_configuration.items_state.items():
        if len(item.ammo_index) != len(state.included_ammo):
            raise InvalidConfiguration(
                "Item {0.name} uses {0.ammo_index} as ammo, but there's only {1} values in included_ammo".format(
                    item, len(state.included_ammo)))

        ammo, locked_ammo = _find_ammo_for(item.ammo_index, ammo_configuration)

        if state.include_copy_in_original_location:
            if item.original_index is None:
                raise InvalidConfiguration(
                    "Item {0.name} does not exist in the original game, cannot use state {1}".format(item, state),
                )
            new_assignment[item.original_index] = create_major_item(item, state, True,
                                                                    resource_database, ammo, locked_ammo)

        for _ in range(state.num_shuffled_pickups):
            item_pool.append(create_major_item(item, state, True, resource_database, ammo, locked_ammo))

        for _ in range(state.num_included_in_starting_items):
            add_resource_gain_to_current_resources(
                create_major_item(item, state, False, resource_database, ammo, locked_ammo).all_resources,
                initial_resources
            )

    return PoolResults(item_pool, new_assignment, initial_resources)



try:
    import PySide2

    has_gui = True
except ModuleNotFoundError as e:
    has_gui = False


def create_subparsers(sub_parsers):
    if has_gui:
        from randovania.gui import qt
        return qt.create_subparsers(sub_parsers)



from enum import Enum


class GameBackendChoice(Enum):
    DOLPHIN = "dolphin"
    NINTENDONT = "nintendont"

    @property
    def pretty_text(self) -> str:
        return _pretty_backend_name[self]


_pretty_backend_name = {
    GameBackendChoice.DOLPHIN: "Dolphin",
    GameBackendChoice.NINTENDONT: "Nintendont",
}


import json
from argparse import ArgumentParser

from randovania import get_data_path
from randovania.layout.preset_migration import VersionedPreset


def refresh_presets_command_logic(args):
    base_path = get_data_path().joinpath("presets")

    with base_path.joinpath("presets.json").open() as presets_file:
        preset_list = json.load(presets_file)["presets"]

    for preset_relative_path in preset_list:
        preset_path = base_path.joinpath(preset_relative_path["path"])
        preset = VersionedPreset.from_file_sync(preset_path)
        preset.ensure_converted()
        preset.save_to_file(preset_path)


def add_refresh_presets_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "refresh-presets",
        help="Loads the preset files and saves then again with the latest version"
    )

    parser.set_defaults(func=refresh_presets_command_logic)



from argparse import ArgumentParser

from randovania.cli.commands.batch_distribute import add_batch_distribute_command
from randovania.cli.commands.distribute import add_distribute_command
from randovania.cli.commands.permalink_command import add_permalink_command
from randovania.cli.commands.randomize_command import add_randomize_command
from randovania.cli.commands.refresh_presets import add_refresh_presets_command
from randovania.cli.commands.validate import add_validate_command

__all__ = ["create_subparsers"]


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "echoes",
        help="Actions regarding Metroid Prime 2: Echoes"
    )
    sub_parsers = parser.add_subparsers(dest="command")
    add_validate_command(sub_parsers)
    add_distribute_command(sub_parsers)
    add_randomize_command(sub_parsers)
    add_batch_distribute_command(sub_parsers)
    add_refresh_presets_command(sub_parsers)
    add_permalink_command(sub_parsers)

    def check_command(args):
        if args.command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



from argparse import ArgumentParser

from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode
from randovania.layout.trick_level import LayoutTrickLevel


def add_debug_argument(parser: ArgumentParser):
    parser.add_argument(
        "--debug",
        choices=range(4),
        type=int,
        default=0,
        help="The level of debug logging to print.")


def add_layout_configuration_arguments(parser: ArgumentParser):
    parser.add_argument(
        "--trick-level",
        type=str,
        choices=[layout.value for layout in LayoutTrickLevel],
        default=LayoutTrickLevel.DISABLED.value,
        help="The level of tricks to use.")
    parser.add_argument(
        "--sky-temple-keys",
        type=str,
        choices=[str(mode.value) for mode in LayoutSkyTempleKeyMode],
        default=LayoutSkyTempleKeyMode.default().value,
        help="The Sky Temple Keys randomization mode.")
    parser.add_argument(
        "--skip-item-loss",
        action="store_true",
        help="Disables the item loss cutscene, disabling losing your items."
    )


def add_validate_argument(parser: ArgumentParser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--validate", action="store_true", dest="validate", default=True,
                       help="After generating a layout, validate if it's possible. Default behaviour.")
    group.add_argument("--no-validate", action="store_false", dest="validate", default=True,
                       help="After generating a layout, don't validate if it's possible.")
